/* Tsinghuaeolus soccer client for Robocup2001
 *
 * Computer Science Department
 * State Key Lab of Intelligent Technology and Systems
 * Tsinghua University
 * 
 * Copyright (C) 2001 Tsinghuaeolus
 * Team Leader : Lishi
 * Team memebers : ChenJiang, YaoJinyi, CaiYunpeng
 *
 * We thank CMU for CMU98 was used for reference in our initial stage.
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message to yjy01@mails.tsinghua.edu.cn.
 * For more information, please see http://www.lits.tsinghua.edu.cn/RoboCup
 */

#ifndef _motion
#define _motion

#include "bpn.h"
#include "types.h"
#include "Command.h"
#include "Strategy.h"
#include "interception.h"
#include "dribble.h"
#include "pass.h"
#include "visualsystem.h"
#include "auditorysystem.h"
#include "shoot.h"
#include "goalie.h"
#include "positioning.h"
#include "clearball.h"
#include "handleball.h"

/******            simple behaviors           *********/

class Behaviors : public Interception, public VisualSystem, public Pass, public Positioning, public AuditoryProcess, public Dribble, public Shoot, public Goalie,  public ClearBall, public Handleball{
private:
	double decided_angle, desire;
public:
	Behaviors();
	void Reset();

/*******************   basic behaviors ***************/
	bool go_to(Vector pos, float rapidness, Command& command);
	bool go_to(Vector pos, float rapidness); // execute at once
	bool smartgoto(Vector pos, float speed);// execute at once
	bool smartgoto(Vector pos, float speed, Command& command);
	AngleDeg turn_to(AngleDeg ang, Command& command);
	AngleDeg turn_to(AngleDeg ang); // execute at once
	AngleDeg turn_to(Vector pos, Command& command);
	AngleDeg turn_to(Vector pos); // execute at once
	bool smartturn(AngleDeg ang, Command& command);
	bool smartturn(AngleDeg ang); // execute at once
	AngleDeg turn_neck_to(AngleDeg ang, AngleDeg bdfy, Command& command);
	AngleDeg turn_neck_to(AngleDeg ang, Command& command);
	AngleDeg turn_neck_to(AngleDeg ang); // execute at once
	bool run(float rapidness, Command& command);
	bool run(float rapidness);  // execute at once
	bool stay(Command& command);
	bool stay();
	bool seize(Command& command);
	bool seize();
	bool donothing(Command& command);
	bool donothing();
	bool circumambulate(Vector pos, float rapidness);
	bool circumambulate(Vector pos, float rapidness, Command& command);
/***************************************************************/
	bool within_range(Vector& pos);

	bool infield(Vector pos); // check if it is in the field
	bool incourt(Vector pos); // check if it is in the court
	friend class laceKick;
};

#endif