/* Tsinghuaeolus soccer client for Robocup2001
 *
 * Computer Science Department
 * State Key Lab of Intelligent Technology and Systems
 * Tsinghua University
 * 
 * Copyright (C) 2001 Tsinghuaeolus
 * Team Leader : Lishi
 * Team memebers : ChenJiang, YaoJinyi, CaiYunpeng
 *
 * We thank CMU for CMU98 was used for reference in our initial stage.
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message to yjy01@mails.tsinghua.edu.cn.
 * For more information, please see http://www.lits.tsinghua.edu.cn/RoboCup
 */

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "Types.h"
#include "Sensory.h"
#include "utils.h"
#include "Global.h"
#include "Params.h"


/*****************************  Sensory  *******************************/
Sensory::Sensory(){
	latestsenseinfo = 0;
	LastSightTime = -1;
	LastSenseTime = -1;
	PrevSenseTime = -1;
	LastSoundTime = -1;
	LastClosestMarker = No_Marker;
	dir_time = -1;
	for(int i = 0; i < CP_num_divisions; i ++){
		dir_conf[i] = 0.0f;
	}
}
/*****************************  Parse  **********************************/
/*	This Parse Module is transgrafted from CMU99 publication code.	We owe special
	thanks to CMU's great effort	*/
void Sensory::Parse(char *SensoryInfo)
{
	SenseType sense_type;
	Time time;

	PrevInterruptTime = LatestTime;
		
	switch ( SensoryInfo[3] ){
	case 'e': sense_type = See_Msg;   break; /* see   */
	case 'n': sense_type = Sense_Msg; break; /* sense */
	case 'r': 
		switch(SensoryInfo[1]){
		case 's': sense_type = SP_Msg;    break; /* server params */
		default: sense_type = No_Msg;			/* maybe (error,warning...) */
		}
		break;	
	case 'a':								 /* hear, player params or player types */
		switch( SensoryInfo[1]){
		case 'h': sense_type = Hear_Msg;  break; /* hear */
		case 'p': /* player params or player types */
			switch(SensoryInfo[8]){
			case 'p': sense_type = PP_Msg; break; /* palyer params */
			case 't': sense_type = PT_Msg; break; /* player types */
			default : sense_type = No_Msg;
			}
			break;
		default: sense_type = No_Msg;
		}
		break;
	case 'o':
		sense_type = SC_Msg; break;
	default:  sense_type = No_Msg;
	}

	switch ( sense_type ){
	/* see */
	case See_Msg:
		// extract time info
		time = get_int(&SensoryInfo);
		PrevSightTime = LastSightTime;
		LastSightTime = time;
		LatestTime = time > LatestTime ? time : LatestTime;

		/* Don't parse a second sight from the a cycle prior to the latest one*/
		if (time <= PrevSightTime) break;/*&& !situation.ClockStopped*///abort sight info when clock stopped
		 
		Parse_Sight(time, SensoryInfo);

		if(TwiceSight) Reset();//too much may result int out of array limit

		if(NewSight) TwiceSight = true;
		else NewSight = true;
    break;

	/* sense */
	case Sense_Msg: 
		// extract time info
		time = get_int(&SensoryInfo);
		
		LatestTime = time > LatestTime ? time : LatestTime;

		Parse_Sense(time, SensoryInfo);
		LastSenseTime = time;
		NewSenseInfo = true;
    break;

	/* hear */
	case Hear_Msg:  
		// extract time info
		time = get_int(&SensoryInfo);
		LatestTime = time > LatestTime ? time : LatestTime;
		Parse_Sound(time, SensoryInfo); 
		LastSoundTime = time;
		NewSound = true;
    break;
	case SP_Msg:
		Parse_ServerParams(SensoryInfo + 13);
	break;
	case PP_Msg:
		Parse_PlayerParams(SensoryInfo +13);
	break;
	case PT_Msg:
		Parse_PlayerTypes(SensoryInfo + 12);
		break;
	case SC_Msg://score
		break;
	}

	LastSenseType = sense_type;
}

void Sensory::Parse_Sense(Time time, char *SenseInfo)
{
	get_word(&SenseInfo);  
	SenseInfo += 10;       /* "view_mode " */
	switch ( SenseInfo[0] )
	{	
	case 'h': ViewQuality = VQ_High; break;  /* high */
	case 'l': ViewQuality = VQ_Low;  break;  /* low  */
	default:  my_error("Unknown view quality");
	}

	LastViewWidth = ViewWidth;
	get_next_word(&SenseInfo);

	switch ( SenseInfo[1] )
	{
	case 'o': ViewWidth = VW_Normal; break;  /* normal */
	case 'a': ViewWidth = VW_Narrow; break;  /* narrow */
	case 'i': ViewWidth = VW_Wide;   break;  /* wide   */
	default:  my_error("Unknown view quality");
	}

	float stamina = get_float(&SenseInfo); /* stamina */
	float effort  = get_float(&SenseInfo); /* effort */
	float speed   = get_float(&SenseInfo); /* speed */
	float rel_spd_angle = get_float(&SenseInfo); /* rel_spd_angle */
	float head_angle = get_float(&SenseInfo); /* head angle */

	int kicks  =   get_int(&SenseInfo); /* kicks */
	int dashes =   get_int(&SenseInfo); /* dashes */
	int turns  =   get_int(&SenseInfo); /* turns */
	int says   =   get_int(&SenseInfo); /* says */
	int turn_necks   =   get_int(&SenseInfo); /* turn_necks */

	SetMySensedInfo(ViewWidth, ViewQuality, stamina,effort,speed,rel_spd_angle,head_angle,kicks,dashes,turns,says,turn_necks,time);
}

#define NOCHNGINFO -500
#define NOFACEINFO -500

void Sensory::Parse_Sight(Time time, char *SightInfo)
{ 
	float dist,ang, dirChng, distChng;
	ObjType object_type;
	char player_side;
	bool IsGoalie ;
	UNum player_number;
	float facedir, neckdir;
	MarkerType marker = No_Marker;
	SideLine  line = SL_No_Line;
	VIEWQUALITY view_qual;
	MarkerType closestMarker = No_Marker;
	bool processThisMarker;
	float closestMarkerDist;

	/* Reset some values */
	num_seenlines.Setdata(0, time);
	num_seenmarkers.Setdata(0, time);
	LastClosestMarker = GetClosestMarker(sensory.PrevSightTime);
	ClosestMarker.Setdata(No_Marker, time);

	while (*SightInfo != ')'){

		dirChng = NOCHNGINFO;
		facedir = NOFACEINFO;
		neckdir = NOFACEINFO;
		player_number = player_side = 0;

		get_word(&SightInfo);        /* " ((" */

		if ( *SightInfo=='g' ){
			
			object_type = OBJ_Marker;/* it is  a "goal " */
			SightInfo += 2;
			if ( *SightInfo=='r' )			marker = Goal_R; 
			else if ( *SightInfo=='l' )     marker = Goal_L; 
			else my_error("goal ?");
		}
		else if ( *SightInfo=='G' ){
			object_type = OBJ_Marker_Behind;
			marker = ClosestGoal();
		} 
		else if ( *SightInfo=='f' ){
			
			object_type = OBJ_Marker;
			SightInfo += 2;          /* it is a "flag " */
			if ( *SightInfo=='r' )
			{
				SightInfo += 2;
				if ( *SightInfo=='0' )		marker = Flag_R0;   
				else if ( *SightInfo=='b' )
				{
					SightInfo += 1;
			    	if ( *SightInfo==')' )  marker = Flag_RB; 
					else
					{
						SightInfo += 1;
						if ( *SightInfo=='1' )		  marker = Flag_RB10; 
						else if ( *SightInfo=='2' )   marker = Flag_RB20; 
						else if ( *SightInfo=='3' )   marker = Flag_RB30; 
				    	else my_error("flag r b ?");
					}
				}
			else if ( *SightInfo=='t' )
			{	
				SightInfo+=1;
				if ( *SightInfo==')' )     marker = Flag_RT; 
				else
				{	SightInfo+=1;
					if ( *SightInfo=='1' )   marker = Flag_RT10; 
					else if ( *SightInfo=='2' )   marker = Flag_RT20; 
					else if ( *SightInfo=='3' )   marker = Flag_RT30; 
					else my_error("flag r t ?");
					}
			}
			else my_error("flag r ?");
			} 
			else if ( *SightInfo=='l' )
			{	
				SightInfo+=2;
				if ( *SightInfo=='0' )       marker = Flag_L0;   
				else if ( *SightInfo=='b' )
				{	
					SightInfo+=1;
			    	if ( *SightInfo==')' )     marker = Flag_LB; 
					else
					{	
						SightInfo+=1;
						if ( *SightInfo=='1' )   marker = Flag_LB10; 
		    			else if ( *SightInfo=='2' )   marker = Flag_LB20; 
			    		else if ( *SightInfo=='3' )   marker = Flag_LB30; 
				    	else my_error("flag l b ?");
					}
				}
				else if ( *SightInfo=='t' )
				{	
					SightInfo+=1;
    				if ( *SightInfo==')' )     marker = Flag_LT; 
	    			else
					{
						SightInfo+=1;
		    			if ( *SightInfo=='1' )   marker = Flag_LT10; 
			    		else if ( *SightInfo=='2' )   marker = Flag_LT20; 
				    	else if ( *SightInfo=='3' )   marker = Flag_LT30; 
				    	else my_error("flag l t ?");
					}  
				} 
				else my_error("flag l ?");
			} 
     		else if ( *SightInfo=='t' )
			{
				SightInfo+=2;
				if ( *SightInfo=='0' )       marker = Flag_T0;   
				else if ( *SightInfo=='l' )
				{
					SightInfo+=2;	
					if ( *SightInfo=='1' )     marker = Flag_TL10; 
					else if ( *SightInfo=='2' )     marker = Flag_TL20; 
					else if ( *SightInfo=='3' )     marker = Flag_TL30; 
					else if ( *SightInfo=='4' )     marker = Flag_TL40; 
					else if ( *SightInfo=='5' )     marker = Flag_TL50; 
					else my_error("flag t l ?");
				}
				else if ( *SightInfo=='r' )
				{	
					SightInfo+=2;
		    		if ( *SightInfo=='1' )     marker = Flag_TR10; 
		        	else if ( *SightInfo=='2' )     marker = Flag_TR20;   
		        	else if ( *SightInfo=='3' )     marker = Flag_TR30; 
		        	else if ( *SightInfo=='4' )     marker = Flag_TR40; 
		        	else if ( *SightInfo=='5' )     marker = Flag_TR50; 
        			else my_error("flag t r ?");
				} 
				else my_error("flag t ?");
			}
			else if ( *SightInfo=='b' )
			{	
				SightInfo+=2;
				if ( *SightInfo=='0' )       marker = Flag_B0;   
				else if ( *SightInfo=='l' )
				{
					SightInfo+=2;
					if ( *SightInfo=='1' )     marker = Flag_BL10; 
					else if ( *SightInfo=='2' )     marker = Flag_BL20;
					else if ( *SightInfo=='3' )     marker = Flag_BL30;
					else if ( *SightInfo=='4' )     marker = Flag_BL40; 
					else if ( *SightInfo=='5' )     marker = Flag_BL50; 
        		    else my_error("flag b l ?");  
				} 
				else if ( *SightInfo=='r' )
				{	
					SightInfo+=2;
				    if ( *SightInfo=='1' )     marker = Flag_BR10; 
			    	else if ( *SightInfo=='2' )     marker = Flag_BR20; 
		    		else if ( *SightInfo=='3' )     marker = Flag_BR30; 
			    	else if ( *SightInfo=='4' )     marker = Flag_BR40; 
			    	else if ( *SightInfo=='5' )     marker = Flag_BR50; 
		    		else my_error("flag b r ?");
				} 
		    	else my_error("flag b ?");
			} 
			else if ( *SightInfo=='c' )	
			{
				SightInfo+=1;
				if ( *SightInfo==')' )       marker = Flag_C;
                else 
				{
					SightInfo+=1;
			    	if ( *SightInfo=='b' )     marker = Flag_CB;
		    		else if ( *SightInfo=='t' )     marker = Flag_CT;
		    		else my_error("flag c ?");
				}
			} 
    		else if ( *SightInfo=='p' )
			{
				SightInfo+=2;
    			if ( *SightInfo=='r' )
				{
					SightInfo+=2;
    				if ( *SightInfo=='t')      marker = Flag_PRT; 
	    			else if ( *SightInfo=='c')      marker = Flag_PRC; 
		    		else if ( *SightInfo=='b')      marker = Flag_PRB; 
			    	else my_error("flag p r ?");
				}
				else if ( *SightInfo=='l' ){

					SightInfo+=2;
			    	if ( *SightInfo=='t')      marker = Flag_PLT; 
			    	else if ( *SightInfo=='c')      marker = Flag_PLC; 
		     		else if ( *SightInfo=='b')      marker = Flag_PLB; 
			    	else my_error("flag p l ?");
				}
		    	else my_error("flag p ?");
			} 
    		else if ( *SightInfo=='g' ){

				SightInfo+=2;
    			if ( *SightInfo=='l' )
				{
					SightInfo+=2;
    				if ( *SightInfo=='t' )     marker = Flag_GLT;
	    			else if ( *SightInfo=='b' )     marker = Flag_GLB;
		    		else my_error("flag g l ?");
				} 
    			else if ( *SightInfo=='r' )
				{
					SightInfo+=2;
    				if ( *SightInfo=='t' )     marker = Flag_GRT;
	    			else if ( *SightInfo=='b' )     marker = Flag_GRB;		  
					else my_error("flag g r ?");
				} 
		    	else my_error("flag g ?");
			} 
     		 else my_error("flag ?");
		} 
		else if ( *SightInfo=='F' ){

			object_type = OBJ_Marker_Behind;
			marker = ClosestFlag(); // could be No_Marker
		}
		else if ( *SightInfo=='l' )
		{
			object_type = OBJ_Line;
			SightInfo+=2;          // "line "
			if ( *SightInfo=='r' ) line   = SL_Right;
			else if ( *SightInfo=='l' )line   = SL_Left;
			else if ( *SightInfo=='t' )line   = SL_Top;
			else if ( *SightInfo=='b' )line   = SL_Bottom;
			else my_error("line ?");	
		}
		else if ( *SightInfo=='p' || *SightInfo=='P' )
		{
			object_type = OBJ_Player;
			SightInfo += 1;            /* it is a "player" */
			if ( *SightInfo == ' ' ) /* there's a team */ 
			{
 				SightInfo += 2;
				if ( !strncmp(SightInfo,MyTeamName,MyTeamNameLen) && *(SightInfo+MyTeamNameLen) == '"'  )
					player_side = situation.MySide;
				else
				{
					if ( TheirTeamName[0] == '\n' )
					{	
						int a=0;
						while ( isalpha(*SightInfo) ) TheirTeamName[a++]=*SightInfo++;
					}
					player_side = situation.TheirSide;
				}
				while ( *SightInfo != ' ' && *SightInfo != ')' ) SightInfo++; /* advance past team name */
				if ( *SightInfo== ' ' )/* there's a number */
				{
					player_number = get_int(&SightInfo);
					if( *SightInfo== ' '){/*It's goalie*/
						get_word(&SightInfo);
						IsGoalie = true;
					}else{
						IsGoalie = false;
					}
				}
			}
		}
		else if ( *SightInfo=='b' || *SightInfo=='B' ) object_type = OBJ_Ball;
		else my_error("unknown object");
		advance_to(')',&SightInfo);              // advance to end of object
	   /************************************/
		ang = get_float(&SightInfo);
		if ( *SightInfo != ')' ) /* 'high' quality     */
		{	
			view_qual = VQ_High;
			dist = ang;
			ang = get_float(&SightInfo);
		}
		else /* 'low ' quality */
		{
			view_qual = VQ_Low;
		}
		if ( view_qual != ViewQuality ) my_error("View quality %d correct?",view_qual);
		if ( *SightInfo != ')' )
		{	
			distChng = get_float(&SightInfo);
			dirChng  = get_float(&SightInfo);
		}
		if ( *SightInfo != ')' )
		{
			if (object_type != OBJ_Player) my_error("Only players should have facedir");
			facedir = get_float(&SightInfo);
			neckdir = get_float(&SightInfo);
		}
		if ( *SightInfo != ')' ) my_error("Should be done with object info here");
		SightInfo++;                                /* ")" */
    /************************************/
		switch (object_type)
		{  
		case OBJ_Marker:
		case OBJ_Marker_Behind:
		  /* Want to save 2 closest for triangulation  */
		  /* don't want marker_behind unless necessary */
		  /* If it was a Marker_Behind and we don't know which one */
		if ( marker == No_Marker )
		{
			if ( object_type != OBJ_Marker_Behind ) my_error("Should know the marker");
				break;	
		}	
		processThisMarker = false;
		if ( view_qual == VQ_Low )
		{               /* low quality   */
				/* DON'T BOTHER PROCESSING ANY??? I don't think it helps ... */
				/* COULD process 2---then triangulate */
				/*if ( closestMarkerDist > 0 ){ */
				/*  closestMarkerDist = 0;  */                 /* Only process 1*/
				/*  processThisMarker = TRUE; */
				/*}*/
				}
		else	/* high quality  */
		{
			if ( closestMarker == No_Marker || dist < closestMarkerDist)
			{
				closestMarker = marker;
				closestMarkerDist = dist;	
				processThisMarker = true;
				ClosestMarker.Setdata(marker, time);
			}
		/* Don't bother with marker motion info --  it from sense_body and my angle 
		if ( dirChng != NOCHNGINFO && dist < motionInfoDist ){
		  motionInfoDist = dist;
		  processThisMarker = TRUE;
		  Mem->ClosestMotionMarker = marker;
		}
		*/
		}
		if ( processThisMarker )
		{
			if ( view_qual == VQ_Low )	SeeMarker(marker, ang, time);// low quality    
			else /* if (dirChng == NOCHNGINFO) */                  /* high quality  */
			
				SeeMarker(marker, dist, ang, time);   /* No motion info*/
		/* else
		  SeeMarker(marker, dist, ang, distChng, dirChng, time); */
		}
		break;
		case OBJ_Line:
			if ( *SightInfo != ')' )	;
		/* There's another line coming.  Assuming lines happen
		   last in the visual string and the closer line comes first */
			else if ( view_qual == VQ_Low )           /* low quality   */
				SeeLine(line, ang, time);
			else                                      /* high quality  */
				SeeLine(line, dist, ang, time);
			break;	
		case OBJ_Ball:	
			if ( view_qual == VQ_Low )                /* low quality   */
				SeeBall(ang, time);
			else if ( dirChng == NOCHNGINFO )                   /* high quality  */
				SeeBall(dist, ang, time);
			else                                      /* know direction*/
				SeeBall(dist, ang, distChng, dirChng, time);
			break;	
		case OBJ_Player:	
			if ( !player_side )    /* Too far for team or num */
			{
				if ( view_qual == VQ_Low )                /* low quality   */
					SeePlayer(ang, time);
				else if ( dirChng == NOCHNGINFO )                   /* high quality  */
					SeePlayer(dist, ang, time);
				else                              /* know direction*/
					my_error("Shouldn't know dirChng when the player's far");
			}  
			else
			{	
				if ( !player_number )  /* Too far for number     */
				{	
					if ( view_qual == VQ_Low )                /* low quality   */
						SeePlayer(player_side, ang, time);
					else if ( dirChng == NOCHNGINFO )                   /* high quality  */
						SeePlayer(player_side, dist, ang, time);
					else                         /* know direction*/
						my_error("Shouldn't know dirChng when the team member's far");
				}
				else	/* know side AND number   */
				{	
					if ( view_qual == VQ_Low )                /* low quality   */
						SeePlayer(player_side, player_number, ang, time, IsGoalie);
					else if ( dirChng == NOCHNGINFO )/* high quality  */
					{	
						printf("%s\n",SightInfo-30);
						my_error("Should know dirChng when know number");
						SeePlayer(player_side, player_number, dist, ang, time,IsGoalie);
					}
					else                            /* know direction*/
						SeePlayer(player_side, player_number, dist, ang, distChng, dirChng, facedir, neckdir, time,IsGoalie);	
				}
			}
		}
	}
}

void Sensory::Parse_Sound(Time time, char *SoundInfo)
{
	if ( SoundInfo[1] == 'r' ){   /* Referee or Coach message */
		SoundInfo += 9;             /* " referee " */
		if (strncmp(SoundInfo, "training", 8) == 0)
			Parse_Trainer_Sound(SoundInfo+8);
		else if (islower(SoundInfo[0]) )
			Parse_Referee_Sound(SoundInfo);
		else
			my_error("Referee sounds should start with lower case letters!");
		return;
	}
    else if ( SoundInfo[1] == 'o' ){  /* Online coach message */
		SoundInfo += 14;                /* online_coach_ */
		if ( SoundInfo[0] == situation.MySide ){
			advance_to(' ',&SoundInfo);   /* advance to end of side */
			SoundInfo++;
			Parse_My_Coach_Sound(time,SoundInfo);
		}
		else if ( SoundInfo[0] == situation.TheirSide ){
			advance_to(' ',&SoundInfo);   /* advance to end of side */
			SoundInfo++;
			Parse_Their_Coach_Sound(time,SoundInfo);
		}
		else 
			my_error("online_coach_?");
		return;
	}
	else if ( SoundInfo[1] == 's'){
		/* Self's message need not process */
	}
	else{
		/* Here you could do something with the heard message */
		motion.ProcessMsg(SoundInfo + 1);
	}
}

void Sensory::Parse_ServerParams(char *SensoryInfo){
	SP_goal_width = get_float(&SensoryInfo);
	SP_inertia_moment = get_float(&SensoryInfo);
	SP_player_size = get_float(&SensoryInfo);
	SP_player_decay = get_float(&SensoryInfo);
	SP_player_rand = get_float(&SensoryInfo);
	SP_player_weight = get_float(&SensoryInfo);
	SP_player_speed_max = get_float(&SensoryInfo);
	SP_paccel_max = get_float(&SensoryInfo);
	SP_stamina_max = get_float(&SensoryInfo);
	SP_stamina_inc = get_float(&SensoryInfo);
	SP_recover_init = get_float(&SensoryInfo);
	SP_recover_dec_thr = get_float(&SensoryInfo);
	SP_recover_min = get_float(&SensoryInfo);
	SP_recover_dec = get_float(&SensoryInfo);
	SP_effort_init = get_float(&SensoryInfo);
	SP_effort_dec_thr = get_float(&SensoryInfo);
	SP_effort_min = get_float(&SensoryInfo);
	SP_effort_dec = get_float(&SensoryInfo);
	SP_effort_inc_thr = get_float(&SensoryInfo);
	SP_effort_inc = get_float(&SensoryInfo);
	SP_kick_rand = get_float(&SensoryInfo);
	SP_team_actuator_noise = get_float(&SensoryInfo);
	SP_prand_factor_l = get_float(&SensoryInfo);
	SP_prand_factor_r = get_float(&SensoryInfo);
	SP_kick_rand_factor_l = get_float(&SensoryInfo);
	SP_kick_rand_factor_r = get_float(&SensoryInfo);
	SP_ball_size = get_float(&SensoryInfo);
	SP_ball_decay = get_float(&SensoryInfo);
	SP_ball_rand = get_float(&SensoryInfo);
	SP_ball_weight = get_float(&SensoryInfo);
	SP_ball_speed_max = get_float(&SensoryInfo);
	SP_baccel_max = get_float(&SensoryInfo);
	SP_dash_power_rate = get_float(&SensoryInfo);
	SP_kick_power_rate = get_float(&SensoryInfo);
	SP_kickable_margin = get_float(&SensoryInfo);
	SP_ctl_radius = get_float(&SensoryInfo); 
	SP_ctl_radius_width = get_float(&SensoryInfo);
	SP_max_power = get_float(&SensoryInfo);
	SP_min_power = get_float(&SensoryInfo);
	SP_max_moment = get_float(&SensoryInfo);
	SP_min_moment = get_float(&SensoryInfo);
	SP_max_neck_moment = get_float(&SensoryInfo);
	SP_min_neck_moment = get_float(&SensoryInfo);
	SP_max_neck_angle = get_float(&SensoryInfo);
	SP_min_neck_angle = get_float(&SensoryInfo);
	SP_visible_angle = get_float(&SensoryInfo);
	SP_feel_distance = get_float(&SensoryInfo);
	SP_wind_dir = get_float(&SensoryInfo);
	SP_wind_force = get_float(&SensoryInfo);
	SP_wind_ang = get_float(&SensoryInfo);
	SP_wind_rand  = get_float(&SensoryInfo);
	SP_kickable_area = get_float(&SensoryInfo);
	SP_catch_area_l = get_float(&SensoryInfo);
	SP_catch_area_w = get_float(&SensoryInfo);
	SP_catch_prob = get_float(&SensoryInfo);
	SP_goalie_max_moves = get_int(&SensoryInfo);
	SP_ckmargin = get_float(&SensoryInfo);
	SP_offside_area = get_float(&SensoryInfo);
	SP_win_no = get_float(&SensoryInfo);
	SP_win_random = get_float(&SensoryInfo);
	SP_say_coach_cnt_max = get_int(&SensoryInfo);
	SP_say_coach_msg_size = get_int(&SensoryInfo);
	SP_clang_win_size = get_int(&SensoryInfo);
	SP_clang_define_win = get_int(&SensoryInfo);
	SP_clang_meta_win = get_int(&SensoryInfo);
	SP_clang_advice_win = get_int(&SensoryInfo);
	SP_clang_info_win = get_int(&SensoryInfo);
	SP_clang_mess_delay = get_int(&SensoryInfo);
	SP_clang_mess_per_cycle = get_int(&SensoryInfo);
	SP_half_time = get_int(&SensoryInfo);
	SP_simulator_step = get_int(&SensoryInfo);
	SP_send_step = get_int(&SensoryInfo);
	SP_recv_step = get_int(&SensoryInfo);
	SP_sense_body_step = get_int(&SensoryInfo);
	SP_lcm_step = get_int(&SensoryInfo);
	SP_say_msg_size = get_int(&SensoryInfo);
	SP_hear_max = get_int(&SensoryInfo);
	SP_hear_inc = get_int(&SensoryInfo);
	SP_hear_decay = get_int(&SensoryInfo);
	SP_catch_ban_cycle = get_int(&SensoryInfo);
	SP_slow_down_factor = get_float(&SensoryInfo);
	SP_use_offside = get_int(&SensoryInfo) == 1;
	SP_forbid_kickoff_offside = get_int(&SensoryInfo) == 1;
	SP_offside_kick_margin = get_float(&SensoryInfo);
	SP_audio_cut_dist = get_float(&SensoryInfo);
	SP_dist_qstep = get_float(&SensoryInfo);
	SP_land_qstep = get_float(&SensoryInfo);
	SP_dir_qstep = get_float(&SensoryInfo);
	SP_dist_qstep_l = get_float(&SensoryInfo);
	SP_dist_qstep_r = get_float(&SensoryInfo);
	SP_land_qstep_l = get_float(&SensoryInfo);
	SP_land_qstep_r = get_float(&SensoryInfo);
	SP_dir_qstep_l = get_float(&SensoryInfo);
	SP_dir_qstep_r = get_float(&SensoryInfo);
	SP_coach_mode = get_int(&SensoryInfo) == 1;
	SP_coach_w_referee_mode = get_int(&SensoryInfo) == 1;
	SP_old_hear = get_int(&SensoryInfo) == 1;
	SP_sv_step = get_int(&SensoryInfo);
}

void Sensory::Parse_PlayerParams(char *SensoryInfo){

}

void Sensory::Parse_PlayerTypes(char *SensoryInfo){
	int id = get_int(&SensoryInfo);
	playertypes[id].player_speed_max = get_float(&SensoryInfo);
	playertypes[id].stamina_inc_max = get_float(&SensoryInfo);
	playertypes[id].player_decay = get_float(&SensoryInfo);
	playertypes[id].inertia_moment = get_float(&SensoryInfo);
	playertypes[id].dash_power_rate = get_float(&SensoryInfo);
	playertypes[id].player_size = get_float(&SensoryInfo);
	playertypes[id].kickable_margin = get_float(&SensoryInfo);
	playertypes[id].kick_rand = get_float(&SensoryInfo);
	playertypes[id].extra_stamina = get_float(&SensoryInfo);
	playertypes[id].effort_max = get_float(&SensoryInfo);
	playertypes[id].effort_min = get_float(&SensoryInfo);
}

void Sensory::Parse_Trainer_Sound(char *msg){
	get_word(&msg);
	if (strncmp(msg, "Prepare_to_train", 16) == 0){
		Prepare_to_train = true;
	}
}

void Sensory::Parse_Referee_Sound(char *msg)
{
    printf("Referee sound: <%s \n", msg);
	msg[strlen(msg) - 2] = 0; /* cut off the newline and paren */
    
	switch( msg[0] ){
	case 'p': situation.HearMode(PM_Play_On); break;           /* play_on */
	case 'k': 
    if ( msg[5] == 'i' ){                                  /* kick_in */
      if ( msg[8] == situation.MySide )
		  situation.HearMode(PM_My_Kick_In);
	  else if ( msg[8] == situation.TheirSide )
		  situation.HearMode(PM_Their_Kick_In);
      else 
		  my_error("kick_in_?");
    }
	else if ( msg[5] == 'o' ){                            /* kick_off */
		if ( msg[9] == situation.MySide )
			situation.HearMode(PM_My_Kick_Off);
		else if ( msg[9] == situation.TheirSide )
			situation.HearMode(PM_Their_Kick_Off);
		else 
			my_error("kick_off_?");
    }
    else
		my_error("referee k..?");
    break;
	case 'g': 
		if ( msg[5] == 'k' ){                                 /* goal_kick */
			if ( msg[10] == situation.MySide )
				situation.HearMode(PM_My_Goal_Kick);
			else if ( msg[10] == situation.TheirSide )
				situation.HearMode(PM_Their_Goal_Kick);
			else 
				my_error("goal_kick_?");
		}
		else if ( msg[5] == 'e' ){                           /* goalie_catch_ball */
			if ( msg[18] == situation.MySide )
				situation.HearMode(PM_My_Goalie_Free_Kick);
			else if ( msg[18] == situation.TheirSide )
				situation.HearMode(PM_Their_Goalie_Free_Kick);
			else
				my_error("goalie_catch_ball_?");
		}
		else if ( msg[5] == situation.MySide ){                    /* goal */
			situation.UpdateScore(situation.MySide);
			situation.SetKickOffMode(KickOff_Theirs);
			situation.HearMode(PM_Before_Kick_Off);
		}
		else if ( msg[5] == situation.TheirSide ){
			situation.UpdateScore(situation.TheirSide);
			situation.SetKickOffMode(KickOff_Mine);
			situation.HearMode(PM_Before_Kick_Off);
		}
		else 
			my_error("referee g..?");
		break;
	case 'c':                                               /* corner_kick */
		if ( msg[12] == situation.MySide )
			situation.HearMode(PM_My_Corner_Kick);
		else if ( msg[12] == situation.TheirSide )
			situation.HearMode(PM_Their_Corner_Kick);
		else 
			my_error("corner_kick_?");
		break;
	case 'd': situation.HearMode(PM_Drop_Ball); break;        /* drop_ball */    
	case 'o':                                               /* offside */    
    if ( msg[8] == situation.MySide )
		situation.HearMode(PM_Their_Offside_Kick);
    else if ( msg[8] == situation.TheirSide )
		situation.HearMode(PM_My_Offside_Kick);
    else 
		my_error("offside_?");
    break;
	case 'f':
		if ( msg[5] == 'k' ){                                 /* free_kick */
			if ( msg[10] == situation.MySide )
				situation.HearMode(PM_My_Free_Kick);
			else if ( msg[10] == situation.TheirSide )
				situation.HearMode(PM_Their_Free_Kick);
			else 
				my_error("free_kick_?");
		}
		else if ( msg[5] == situation.MySide )                     /* foul */
      ;
    else if ( msg[5] == situation.TheirSide )
      ;
    else 
		my_error("referee f..?");
    break;
	case 'h':                                               /* half_time */
		situation.HearMode(PM_Half_Time);  /* play_mode to before_kick_off        */
		if ( situation.MySide == 'l' )
			situation.SetKickOffMode(KickOff_Theirs);
		else 
			situation.SetKickOffMode(KickOff_Mine);
		break;             
	case 'b': situation.HearMode(PM_Before_Kick_Off); break;       /* before_kick_off */
	case 't': 
		if ( msg[5] == 'u' ){                             /* time_up */
			situation.HearMode(PM_Time_Up); 
		}
		else if ( msg[5] == 'o' )                             /* time_over */
		{
			break;
		}
		else if ( msg[5] == 'e' ){                            /* time_extended */
			situation.HearMode(PM_Extended_Time);
			if ( situation.MySide == 'l' )
				situation.SetKickOffMode(KickOff_Mine);
			else 
				situation.SetKickOffMode(KickOff_Theirs);
		} 
		else 
			my_error("referee t..?");
		break;
	default: my_error("Referee msg ????");
    }
}

void Sensory::Parse_initialize_message(char *recvbuf){
	char mode[100];
	if ( !(strncmp(recvbuf,"(init",4)) ) {
    /* It's an init msg */
		sscanf(recvbuf,"(init %c %d %[^)]",&situation.MySide, &MyNumber, mode);
		if (situation.MySide == 'r'){
			situation.TheirSide = 'l';
			MyGoal = Goal_R;
			TheirGoal = Goal_L;
		}
		else{
			situation.TheirSide = 'r';
			MyGoal = Goal_L;
			TheirGoal = Goal_R;
		}

		ServerAlive = 1;
	}
	else if ( !(strncmp(recvbuf,"(reconnect",4)) ) {
    /* It's a reconnect msg */
		sscanf(recvbuf,"(reconnect %c %[^)]",&situation.MySide, mode);
		MyNumber = IP_reconnect;
		printf("reconnecting to %d on side %c!\n",MyNumber,situation.MySide);
		if (situation.MySide == 'r'){
			situation.TheirSide = 'l';
			MyGoal = Goal_R;
			TheirGoal = Goal_L;
		}
		else{
			situation.TheirSide = 'r';
			MyGoal = Goal_L;
			TheirGoal = Goal_R;
		}

		ServerAlive = 1;
	}
	else {
		my_error("Didn't get an init message: '%s')",recvbuf);
		ServerAlive = 0;
		PostMessage(hWND, WM_DESTROY, 0, 0);
	}

	if ( CP_goalie && FP_goalie_number != MyNumber )
		my_error("goalie number inconsistent with me being goalie");

	if ( !CP_goalie && FP_goalie_number == MyNumber )
		my_error("I should be the goalie");

	if ( mode[0] == 'b' ){ /* Before_kick_off */ 
		situation.HearMode(PM_Before_Kick_Off);
		if ( situation.MySide == 'l' )
			situation.SetKickOffMode(KickOff_Mine);
		else 
			situation.SetKickOffMode(KickOff_Theirs);
	}
	else                  /* Act as if the game's in progress */
		situation.HearMode(PM_Play_On);
}

/********************Set Sensed Infomation******************************************/

void Sensory::SetMySensedInfo(VIEWWIDTH viewidth, VIEWQUALITY viewqual, float stamina,float effort,float speed,AngleDeg rel_spd_angle,
	AngleDeg head_angle,int kicks,int dashes,int turns,int says,int turn_necks,Time time){
	if ( LastSenseTime == time && !situation.ClockStopped)
		return;

	PrevSenseTime = LastSenseTime;
	LastSenseTime = time;

	latestsenseinfo = !latestsenseinfo;
	sensebodyinfo[latestsenseinfo].setvalues(viewidth, viewqual, stamina,effort,speed,rel_spd_angle,head_angle,kicks,dashes,
							                  turns,says,turn_necks,time);
}

SenseBodyInfo& Sensory::GetSenseBodyInfo(Time time){
	return (time == LastSenseTime) ? sensebodyinfo[latestsenseinfo] : sensebodyinfo[!latestsenseinfo];
}

/************************	See Module	***************************/
void Sensory::SeeMarker(MarkerType marker,AngleDeg ang,Time time){
	//there are so many marker, foget those without known distance
	//this rem is non-trivial, because update assume all markers's distance is known

	//Markers[marker].set_polar(ang,time);
	//SeenMarkers.Data(time)[[num_seenmarkers.Data(time)++]] = marker;
}

void Sensory::SeeMarker(MarkerType marker,float dist,AngleDeg ang,Time time){
	Markers[marker].set_polar(ang,dist,time);
	SeenMarkers.Setdata(marker, num_seenmarkers.Data(time)++, time);
}

void Sensory::SeeLine(SideLine line,AngleDeg ang,Time time){
	Lines[line].set_polar(ang,time);
	SeenLines.Setdata(line, num_seenlines.Data(time)++, time);
}

void Sensory::SeeLine(SideLine line ,float dist,AngleDeg ang,Time time){
	Lines[line].set_polar(ang,dist,time);
	SeenLines.Setdata(line, num_seenlines.Data(time)++, time);
}

void Sensory::SeeBall(AngleDeg ang,Time time){
	ball.set_polar(ang,time);
}

void Sensory::SeeBall(float dist,AngleDeg ang,Time time){
	ball.set_polar(ang,dist,time);
}

void Sensory::SeeBall(float dist,AngleDeg ang,float distChng,float dirChng,Time time){
	ball.set_polar(ang,dist,time);
	ball.set_chinfo(distChng,dirChng,time);
}

void Sensory::SeePlayer(AngleDeg ang,Time time){
	if (num_teamlessplayers >= SP_team_size)
		return;
	TeamlessPlayers[num_teamlessplayers ++].set_polar(ang, time);
}

void Sensory::SeePlayer(float dist, AngleDeg ang,Time time){
	if (num_teamlessplayers >= SP_team_size)
		return;
	TeamlessPlayers[num_teamlessplayers ++].set_polar(ang, dist, time);
}

void Sensory::SeePlayer(char player_side, AngleDeg ang,Time time){
	if(player_side == situation.MySide){
		if (num_unknown_myplayers >= SP_team_size)
			return;
		UnknownMyPlayers[num_unknown_myplayers ++].set_polar(ang, time);
	}
	else{
		if (num_unknown_theirplayers >= SP_team_size)
			return;
		UnknownTheirPlayers[num_unknown_theirplayers ++].set_polar(ang, time);
	}
}

void Sensory::SeePlayer(char player_side,float dist,AngleDeg ang,Time time){
	if(player_side == situation.MySide){
		if (num_unknown_myplayers >= SP_team_size)
			return;
		UnknownMyPlayers[num_unknown_myplayers ++].set_polar(ang, dist, time);
	}
	else{
		if (num_unknown_theirplayers >= SP_team_size)
			return;
		UnknownTheirPlayers[num_unknown_theirplayers ++].set_polar(ang, dist, time);
	}
}

void Sensory::SeePlayer(char player_side,UNum NO,AngleDeg ang,Time time,bool IsGoalie){
	if(player_side == situation.MySide){
		MyPlayer(NO).set_polar(ang,time);
		IdentifyMyGoalie(NO, IsGoalie);
	}
	else{
		TheirPlayer(NO).set_polar(ang,time);
		IdentifyTheirGoalie(NO, IsGoalie);
	}
}

void Sensory::SeePlayer(char player_side,UNum NO,float dist,AngleDeg ang,Time time,bool IsGoalie){
	if(player_side == situation.MySide){
		MyPlayer(NO).set_polar(ang,dist,time);		
		IdentifyMyGoalie(NO, IsGoalie);
	}
	else{
		TheirPlayer(NO).set_polar(ang,dist,time);		
		IdentifyTheirGoalie(NO, IsGoalie);
	}
}

void Sensory::SeePlayer(char player_side,UNum NO,float dist,AngleDeg ang,float distChng,
	float dirChng,AngleDeg facedir,AngleDeg neckdir,Time time,bool IsGoalie){
	if(player_side == situation.MySide){
		MyPlayer(NO).set_polar(ang,dist,time);
		MyPlayer(NO).set_chinfo(distChng,dirChng,time);
		MyPlayer(NO).set_fcinfo(facedir, neckdir,time);	
		IdentifyMyGoalie(NO, IsGoalie);
	}
	else{
		TheirPlayer(NO).set_polar(ang,dist,time);
		TheirPlayer(NO).set_chinfo(distChng,dirChng,time);
		TheirPlayer(NO).set_fcinfo(facedir, neckdir,time);
		IdentifyTheirGoalie(NO, IsGoalie);
	}
}

void Sensory::IdentifyMyGoalie(UNum No,bool IsGoalie){
	MyPlayer(No).Is_goalie = IsGoalie;
	if(IsGoalie) fieldinfo.mygoalie = No;
} 

void Sensory::IdentifyTheirGoalie(UNum No,bool IsGoalie){
	TheirPlayer(No).Is_goalie = IsGoalie;
	if(IsGoalie) fieldinfo.theirgoalie = No;
}
/*******************************************************************************/

void Sensory::Update(){

	situation.UpdateTime(LatestTime) ;

	if ( NewSight && LastSightTime < situation.CurrentTime - 1 && !situation.ClockStopped){
    /* Special case--ignore sights from just before change to play_on_mode
       if they seem to be 2 cycles ago.  Better would be to adjust the sight
       time, but then sight times of all other objects also need to be adjusted */
		NewSight = false; /* process only recent cycles */
		if ( LastSightTime < situation.CurrentTime - 2 )
			my_error("Last sight shouldn't be so out of date");
	}
	
	action.ActionsInvalidation(); // is the last cycle actions taking effect?
	Self.update(situation.CurrentTime);
	Update_others(situation.CurrentTime);

	Update_dirconf(situation.CurrentTime, NewSight);
	action.PreProcess();

	if (situation.ClockStopped)
		situation.ClockStoppedTime ++;
	Reset();
	situation.SightTime = sensory.LastSightTime;

	situation.UpdateMode();
}

void Sensory::Update_dirconf(Time time, bool NewSight){
	for(int i = 0; i < CP_num_divisions; i ++){
		dir_conf[i] *= Exp(CP_conf_decay, (float)(time - dir_time));
	}
	if (NewSight){
		for(i = 0; i < CP_num_divisions; i ++){
			if (fabs(NormalizeAngle((i + 0.5f) * CP_division - Self.Headfacing(LastSightTime))) < 0.8f * sensory.MyViewAngle()){
				dir_conf[i] = CP_conf_max;
				if (LastSightTime != time)
					dir_conf[i] *= Exp(CP_conf_decay, (float)(time -LastSightTime));
			}
		}
	}
	dir_time = time;
}

void Sensory::Reset(){
	NewSight = false;
	NewSenseInfo = false;
	NewSound = false;
	TwiceSight = false;
	num_unknown_myplayers = 0;
	num_unknown_theirplayers = 0;
	num_teamlessplayers = 0;
}

float Sensory::DirConf(AngleDeg dir){
	return dir_conf[int(NormalizeAngle(dir, 0.0f) / CP_division)];
}

void Sensory::Update_others(Time time){
	if(motion.Vinfo_coming){
		if(!sensory.NewSight){
			DoLog(LOG_VDEC,"coming visual haven't come");
		}else{
		}
	}

	ball.update(time);// ball info should be updated prior to the other objs
	int i,j;
	for(i = 0; i < SP_team_size; i ++){
		MyTeam[i].update(time);
	}

	for(i = 0; i < SP_team_size; i ++){
		TheirTeam[i].update(time);
	}

	for(i = 0; i < num_unknown_theirplayers; i ++){
		if (!UnknownTheirPlayers[i].IsDistanceSeen()) continue;
		UnknownTheirPlayers[i].Reset(UnknownTheirPlayers[i].seen_time());
		UnknownTheirPlayers[i].update(time);
	}
	for(i = 0; i < num_unknown_myplayers; i ++){
		if (!UnknownMyPlayers[i].IsDistanceSeen()) continue;
		UnknownMyPlayers[i].Reset(UnknownMyPlayers[i].seen_time());
		UnknownMyPlayers[i].update(time);
	}
	for(i = 0; i < num_teamlessplayers; i ++){
		if (!TeamlessPlayers[i].IsDistanceSeen()) continue;		
		TeamlessPlayers[i].Reset(TeamlessPlayers[i].seen_time());
		TeamlessPlayers[i].update(time); 
	}
	
	unknowntheirplayerlist.update(time);
	unknownmyplayerlist.update(time);
	teamlessplayerlist.update(time);
	int idx;
	float dist, min_dist;

	if(sensory.NewSight ){
	/*	Far Player Update	*/
		Vector pos;
		for(i = num_unknown_theirplayers-1; i >= 0; i--){//descent order to reduce update when twice sight
			if (!UnknownTheirPlayers[i].IsDistanceSeen()) continue;
			idx = -1;
			min_dist = CP_too_different_distance;
			for(int j = 0; j < SP_team_size; j ++){
				if(TheirTeam[j].IsSeen(UnknownTheirPlayers[i].seen_time())) continue;
				if(TheirTeam[j].pos_conf > CP_identify_min_conf){
					dist = TheirTeam[j].pos.disaeolus(UnknownTheirPlayers[i].pos);
					if (dist < min_dist){
						idx = j;
						min_dist = dist;
					}
				}
			}
			if(idx != -1 ){
				if(TheirTeam[idx].IsSamePlayer(UnknownTheirPlayers[i])){
					if(TheirTeam[idx].seen_time() < UnknownTheirPlayers[i].seen_time()){
						TheirTeam[idx].set_polar(UnknownTheirPlayers[i].seen_rel_angle(), UnknownTheirPlayers[i].seen_distance(),UnknownTheirPlayers[i].seen_time());
						TheirTeam[idx].update(time);
					}else{
						//twice sight don't need to update;
					}
					DoLog(LOG_UNKPLAYER,"Findnum Opp %.0f pos(%.2f,%.2f)",(float)(idx+1),TheirTeam[idx].pos.x,TheirTeam[idx].pos.y);
					continue;
				}
			}
			//	Enlist in the unknownplayer list
			unknowntheirplayerlist.EnList(UnknownTheirPlayers[i]);
		}

		for(i = num_unknown_myplayers-1; i >= 0; i--){
			if (!UnknownMyPlayers[i].IsDistanceSeen()) continue;
			DoLog(LOG_UNKPLAYER,"find teammate for(%.1f,%.1f) at %d", UnknownMyPlayers[i].pos.x, UnknownMyPlayers[i].pos.y, UnknownMyPlayers[i].seen_time());
			idx = -1;
			min_dist = CP_too_different_distance;
			for(int j = 0; j < SP_team_size; j ++){
				if(MyTeam[j].IsSeen(UnknownMyPlayers[i].seen_time())) continue;
				if(MyTeam[j].pos_conf > CP_identify_min_conf){
					dist = MyTeam[j].pos.disaeolus(UnknownMyPlayers[i].pos);
					//DoLog(LOG_UNKPLAYER,"to %d(%.1f,%.1f), dis %.1f",MyTeam[j].InsideNO, MyTeam[j].Pos(seentime).x, MyTeam[j].Pos(seentime).y,dist);
					if (dist < min_dist){
						idx = j;
						min_dist = dist;
					}
				}else{
				//	DoLog(LOG_UNKPLAYER,"omit %d for conf %.1f", MyTeam[j].InsideNO,MyTeam[j].pos_conf); 
				}
			}
			
			if(idx != -1 ){
				if(MyTeam[idx].IsSamePlayer(UnknownMyPlayers[i])){
					if(MyTeam[idx].seen_time() < UnknownMyPlayers[i].seen_time()){
						MyTeam[idx].set_polar(UnknownMyPlayers[i].seen_rel_angle(), UnknownMyPlayers[i].seen_distance(),UnknownMyPlayers[i].seen_time());
						MyTeam[idx].update(time);
					}
					DoLog(LOG_UNKPLAYER,"Findnum Teammate %.0f pos(%.2f,%.2f)",(float)(idx+1),MyTeam[idx].pos.x,MyTeam[idx].pos.y);
					continue;
				}
			}
			//	Enlist in the unknownplayer list
			unknownmyplayerlist.EnList(UnknownMyPlayers[i]);
		}
	/*	Feeled Player Update	*/
		for(i = num_teamlessplayers-1; i >= 0; i--){
			if (!TeamlessPlayers[i].IsDistanceSeen()) continue;

			idx = -1;
			min_dist = CP_too_different_distance;
			if (TeamlessPlayers[i].seen_distance() > SP_feel_distance) continue;
			
			for(j = 0; j < SP_team_size; j ++){
				if(TheirTeam[j].IsSeen(TeamlessPlayers[i].seen_time())) continue;
				if(TheirTeam[j].pos_conf < CP_identify_min_conf) continue;
				dist = TheirTeam[j].pos.disaeolus(TeamlessPlayers[i].pos);
				if (dist < min_dist){
					idx = j + SP_team_size;
					min_dist = dist;
				}
			}
			
			for(j = 0; j < SP_team_size; j ++){
				if(MyTeam[j].IsSeen(TeamlessPlayers[i].seen_time())) continue;
				if(MyTeam[j].pos_conf < CP_identify_min_conf) continue;
				dist = MyTeam[j].pos.disaeolus(TeamlessPlayers[i].pos);
				if (dist < min_dist){
					idx = j;
					min_dist = dist;
				}
			}
			
			if (idx != -1 && min_dist < CP_too_different_distance){
				if (idx >= 11){
					idx -= SP_team_size;
					if(TheirTeam[idx].IsSamePlayer(TeamlessPlayers[i])){
						if(TheirTeam[idx].seen_time() < TeamlessPlayers[i].seen_time()){
							TheirTeam[idx].set_polar(TeamlessPlayers[i].seen_rel_angle(), TeamlessPlayers[i].seen_distance(), TeamlessPlayers[i].seen_time());
							TheirTeam[idx].update(time);
						}
						DoLog(LOG_UNKPLAYER,"Identify Oppnum %d pos(%.2f,%.2f)",idx+1, TheirTeam[idx].pos.x, TheirTeam[idx].pos.y);
						continue;
					}
				}else{
					if(MyTeam[idx].IsSamePlayer(TeamlessPlayers[i])){
						if(MyTeam[idx].seen_time() < TeamlessPlayers[i].seen_time()){
							MyTeam[idx].set_polar(TeamlessPlayers[i].seen_rel_angle(), TeamlessPlayers[i].seen_distance(),TeamlessPlayers[i].seen_time());
							MyTeam[idx].update(time);
						}
						DoLog(LOG_UNKPLAYER,"Identify Teammatenum%.0f pos(%.2f,%.2f)",(float)(idx+1),MyTeam[idx].pos.x,MyTeam[idx].pos.y);
						continue;
					}
				}
			}
			
			idx = -1;
			min_dist = CP_too_different_distance;
			for(j = 0; j < unknowntheirplayerlist.oldnumplayers; j++){
				if(unknowntheirplayerlist.List[j]->IsSeen(TeamlessPlayers[i].seen_time())) continue;
				dist = unknowntheirplayerlist.List[j]->pos.disaeolus(TeamlessPlayers[i].pos);
				if(dist < min_dist){
					idx = j + SP_team_size;
					min_dist = dist;
				}
			}
			for(j = 0; j < unknownmyplayerlist.oldnumplayers; j++){
				if(unknownmyplayerlist.List[j]->IsSeen(TeamlessPlayers[i].seen_time())) continue;
				dist = unknownmyplayerlist.List[j]->pos.disaeolus(TeamlessPlayers[i].pos);
				if(dist < min_dist){
					idx = j;
					min_dist = dist;
				}
			}
			if (idx != -1 && min_dist < CP_different_distance){
				if (idx >= SP_team_size){
					idx -= SP_team_size;
					if(unknowntheirplayerlist.List[idx]->IsSamePlayer(TeamlessPlayers[i])){
						if(unknowntheirplayerlist.List[idx]->seen_time() < TeamlessPlayers[i].seen_time()){
							unknowntheirplayerlist.List[idx]->set_polar(TeamlessPlayers[i].seen_rel_angle(), TeamlessPlayers[i].seen_distance(),TeamlessPlayers[i].seen_time());
							unknowntheirplayerlist.List[idx]->update(time);
						}
						DoLog(LOG_UNKPLAYER,"Find Opp (%.2f,%.2f) at %d",unknowntheirplayerlist.List[idx]->pos.x,unknowntheirplayerlist.List[idx]->pos.y,TeamlessPlayers[i].seen_time());
						continue;
					}
				}else{
					if(unknownmyplayerlist.List[idx]->IsSamePlayer(TeamlessPlayers[i])){
						if(unknownmyplayerlist.List[idx]->seen_time() < TeamlessPlayers[i].seen_time()){
							unknownmyplayerlist.List[idx]->set_polar(TeamlessPlayers[i].seen_rel_angle(), TeamlessPlayers[i].seen_distance(),TeamlessPlayers[i].seen_time());
							unknownmyplayerlist.List[idx]->update(time);
						}
						DoLog(LOG_UNKPLAYER,"Find Teammate (%.2f,%.2f) at %d",unknownmyplayerlist.List[idx]->pos.x,unknownmyplayerlist.List[idx]->pos.y,TeamlessPlayers[i].seen_time());
						continue;
					}
				}
			}
			teamlessplayerlist.EnList(TeamlessPlayers[i]);
		}
	}

	/***********************update end********************/
	ball.update_end(); 
	for(i = 0; i < SP_team_size; i ++){
		MyTeam[i].update_end();
	}
	for(i = 0; i < SP_team_size; i ++){
		TheirTeam[i].update_end();
	}

	unknowntheirplayerlist.update_end();
	unknownmyplayerlist.update_end();
	teamlessplayerlist.update_end();
	
	unknownmyplayerlist.LogList();
	unknowntheirplayerlist.LogList();
	teamlessplayerlist.LogList();
}

int Sensory::MySightInterval(){  
	int interval = SP_send_step;  
	if ( ViewWidth == VW_Narrow ) interval /= 2;  
	else if ( ViewWidth == VW_Wide ) interval *= 2;
	if ( ViewQuality == VQ_Low ) interval /=2;  
	return interval;
}

AngleDeg Sensory::MyViewWidth(VIEWWIDTH view_width){
	switch(view_width){
	case VW_Normal : return 90;
	case VW_Narrow : return 45;
	case VW_Wide   : return 180;
	}
	return 0.0f;
}

AngleDeg Sensory::MyViewAngle(VIEWWIDTH view_width){
	return MyViewWidth(view_width) / 2.0f;
}

AngleDeg Sensory::MyViewWidth(VWTime time){
	VIEWWIDTH VWidth = (time == VWT_Current) ? ViewWidth : NextViewWidth;
	return MyViewWidth(VWidth);
}


AngleDeg Sensory::MyViewAngle(VWTime time){
	return MyViewWidth(time) / 2.0f;
}

MarkerType Sensory::ClosestGoal(){
	if (Self.pos.x < 0)
		return MyGoal;
	else
		return TheirGoal;
}

MarkerType Sensory::ClosestFlag(){
	Vector gap = Self.pos - Markers[LastClosestMarker].pos;
	if (gap.mod() > 1)
		return No_Marker;
	else
		return LastClosestMarker;
}