/* Tsinghuaeolus soccer client for Robocup2001
 *
 * Computer Science Department
 * State Key Lab of Intelligent Technology and Systems
 * Tsinghua University
 * 
 * Copyright (C) 2001 Tsinghuaeolus
 * Team Leader : Lishi
 * Team memebers : ChenJiang, YaoJinyi, CaiYunpeng
 *
 * We thank CMU for CMU98 was used for reference in our initial stage.
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message to yjy01@mails.tsinghua.edu.cn.
 * For more information, please see http://www.lits.tsinghua.edu.cn/RoboCup
 */

#include "goalie.h"
#include "bpn.h"
#include "types.h"
#include "Command.h"
#include "Strategy.h"
#include "global.h"
#include "Utils.h"

/**************************goalie***************************/
Goalie::Goalie(){

}

bool Goalie::Smartgoalie(){
	num_PIT_infos = 0;
	bottomline.LineFromTwoPoints(Vector(CP_safecatch_width - SP_semi_pitch_length,0),Vector(CP_safecatch_width - SP_semi_pitch_length,1));

	if (ball.distance>SP_catch_area_l
		||motion.TheirPlayer_Close2Ball(0).pos.dist(ball.pos)<7
		||(!ball.kickable()&&ball.speed_valid()&&
				(ball.global_vel.x<0||ball.speed<0.9f*CP_goalie_kick_min_vel)))
		GoalieJustKicked = false;
	if(!Ifcare()){
		DoLog(LOG_GOALIE,"Ball far away");
		Dawdle();
		return true;
	}
	
	
	if(ball.catchable()){
		if (!GoalieJustKicked)
		{
			if (!CatchWait())
			{
				Catchball();
				mediator.SetViewWidth(VW_Narrow); 
				return true;
			}
			else
			{
				if (OneCycleIT())
					return true;
				else
				{
					DoLog(LOG_BUG,"wait but no interception!");
					return false;
				}
			}
		}
		else DoLog(LOG_GOALIE,"Should not catch");
	}
	
	
	if(ball.kickable()){
		Kickaway();
		return true;
	}

	if (ball.pos_valid() &&ball.speed_valid() && GetInterceptionInfo(ball.pos,ball.global_vel,Self.pos,Self.global_vel,Self.bodyfacing))
		minITcycles = PIT_info[0].IT_cycles;
	else
		minITcycles = InfCycles;
	DoLog(LOG_GOALIE,"num_PIT %d",num_PIT_infos);
	bool intercept = false;	
	if (true){
		intercept =	GoalieIntercept();
	}
	if (!intercept) DefendMove();
	return true;
} 

bool Goalie::GoalieDefense(){
	GoalieJustKicked = false;
	if(!Ifcare()){
		Dawdle();
		return true;
	}
	bottomline.LineFromTwoPoints(Vector(CP_safecatch_width - SP_semi_pitch_length,0),Vector(CP_safecatch_width - SP_semi_pitch_length,1));
	
	DefendMove();
	
	return true;
}

bool Goalie::Ifcare(){
	return (ball.pos.x < 40 - SP_semi_pitch_length || ball.distance < 30);
} 
bool Goalie::IfOut(){
	if(!ball.pos_valid()) return false;
	if (IsMyPlayer(situation.ballcontroller) && ball.distance>3.5f)
		return false;
	bool controller=bool(situation.most_promising_controller==FP_goalie_number);
	float Teammatecyclegap;
	if(controller){
		Teammatecyclegap = motion.MyPlayer_CtrlOfBall(1).IT_inf.IT_cycles - motion.TheirPlayer_CtrlOfBall(0).IT_inf.IT_cycles;		
	}else{
		Teammatecyclegap = motion.MyPlayer_CtrlOfBall(0).IT_inf.IT_cycles - motion.TheirPlayer_CtrlOfBall(0).IT_inf.IT_cycles;
	}

	if(minITcycles <3) return true;
	bool withforbiddenarea;
	withforbiddenarea = (num_PIT_infos == 0) ? fieldinfo.my_penaltyarea.IsWithin(Self.IT_inf.IT_point) : fieldinfo.my_penaltyarea.IsWithin(PIT_info[0].IT_point);
	if(minITcycles<motion.TheirPlayer_CtrlOfBall(0).IT_inf.IT_cycles ){
		if(Teammatecyclegap>=0 || withforbiddenarea)
			return true;
		else
		{
			DoLog(LOG_GOALIE,"Not out intercepting");
			return false;
		}
	}else
		return false;	
}
bool Goalie::IfOppkickmore(){
	//ball is under control of opponent
	Vector p;
	Ray ballray(ball.pos,ball.vel_angle);
	if (!fieldinfo.SideLines[SL_Left].RayIntersection(ballray,p))
		return true;
	return bool( ballrollingtime(ball.speed,p.dist(ball.pos)) > motion.TheirPlayer_CtrlOfBall(0).IT_inf.IT_cycles);
} 
void Goalie::Dawdle(){
	GoalieRunto(GetGuardPoint(),0,CP_setplay_speed);
}

void Goalie::DefendMove(){
	bool shouldrun = false;
	if (!ball.pos_valid())
		return;
	
	if (!situation.BallFree || !ball.speed_valid())
		ballpoint = ball.pos;
	else
	// need further consideration
		ballpoint = ball.PredictPos((int)situation.min_IT_cycles);

	if (ballpoint.x < -SP_semi_pitch_length)
		ballpoint.x = -SP_semi_pitch_length;
	closerpost = ballpoint.y < 0 ?	fieldinfo.l_mygoalpost : fieldinfo.r_mygoalpost;
	DEF_info[0].OppPos = ballpoint;
	CalcToleranceTime(DEF_info[0]);
	DoLog(LOG_GOALIE, "Tolarence %.1f",DEF_info[0].TolerantTime);
	num_opp_considered = 1;
		
	Defendline.SetValues(ballpoint,MidAngle2((fieldinfo.l_mygoalpost - ballpoint).Angle(), 
			(fieldinfo.r_mygoalpost - ballpoint).Angle()));
	DoLog(LOG_GOALIE,"Defend Ray (%.1f %.1f) %.1f",ballpoint.x,ballpoint.y,MidAngle2((fieldinfo.l_mygoalpost - ballpoint).Angle(),(fieldinfo.r_mygoalpost - ballpoint).Angle())); 

	if (!DefendShootOK()){
		DoLog(LOG_GOALIE,"Def Shoot");
		shouldrun = DefendShoot();
	}
	else
	{
		DoLog(LOG_GOALIE,"Def Pass");
		shouldrun = DefendPass();
	}
	if (!shouldrun)
		AdjustFacings();
}


float Goalie::GetGothroughSpeed(Ray ballcourse,Vector Playerpos, Line pline){
	//assume that the player is always before the target line
	Vector secpoint;
	Vector origin = ballcourse.GetPoint(0.0f);
	if (!ballcourse.intersection(pline,secpoint))
		return CP_impossible_speed;
	float t = ballcourse.DistanceFromOrigin(secpoint);
	Vector rpos = Polar2Vector(secpoint.dist(origin),ballcourse.StrechAngle(secpoint));
	if (fieldinfo.my_penaltyarea.IsWithin(Playerpos))
		return calc_goalie_goingthroughspeed(rpos,t);
	else
		return calc_goingthroughspeed(rpos,t);
}

bool Goalie::CheckShootSpeed(Vector BallPoint,Vector DefPos ){
	float langle = (fieldinfo.l_mygoalpost -BallPoint).Angle();
	float rangle = (fieldinfo.r_mygoalpost -BallPoint).Angle();
	L_through_spd = GetGothroughSpeed(Ray(BallPoint + Polar2Vector(1.5f,langle),langle),DefPos,fieldinfo.SideLines[SL_Left]);
	R_through_spd = GetGothroughSpeed(Ray(BallPoint + Polar2Vector(1.5f,rangle),rangle),DefPos,fieldinfo.SideLines[SL_Left]);
	return (L_through_spd > SP_ball_speed_max + 0.15f && R_through_spd > SP_ball_speed_max + 0.15f);

}

bool Goalie::DefendShootOK(){
	 return CheckShootSpeed(ballpoint,Self.pos) && CheckShootSpeed(ballpoint,Self.pos);
}

bool Goalie::DefendShoot(){
	Vector defendpoint;
		fieldinfo.my_goaldefendbox.RayIntersection(Defendline,defendpoint);
	if (defendpoint.dist(Self.pos) < 3.0f)
	{
		Vector ppoint;
		if (Defendline.intersection(Line(Ray(Self.PredictPos(),Self.bodyfacing)),ppoint))
			defendpoint = ppoint;
	}
	if (defendpoint.dist(ballpoint)< CP_safecatch_width)
		defendpoint = Defendline.GetPoint(CP_safecatch_width);
	if (!fieldinfo.my_penaltyarea.IsWithin(defendpoint))
		fieldinfo.my_penaltyarea.RayIntersection(Defendline,defendpoint);	
	Vector defpoints;
	int mincycles = (int)InfCycles;
	int thiscycles;
	float pointgap = Defendline.intersection(bottomline)/8;
	float angledif;
	Vector gap;
	for (int i=0;i<8;i++)
	{
		defpoints = Defendline.GetPoint(pointgap *(i+1));
		if(fieldinfo.my_penaltyarea.IsWithin(defpoints) && CheckShootSpeed(ballpoint,defpoints))
		{
			gap =defpoints - Self.pos;
			if (gap.mod() < ball.distance/20) continue;
			angledif =(float) fabs(NormalizeAngle(gap.Angle() - Self.bodyfacing));
			if (angledif > 90) angledif = 180 - angledif;
			if (fabs(gap.mod() * angledif) < SP_kickable_margin)
				thiscycles = 0;
			else if (angledif < action.Turnrate(Self.speed))
				thiscycles = 1;
			else
				thiscycles = 2;
			thiscycles += int(gap.mod()/SP_player_speed_max - ballrollingtime(SP_player_speed_max,pointgap * (i+1))); 
			if (thiscycles  < mincycles)
			{
				mincycles = thiscycles;
				defendpoint = defpoints;
			}
		}			
	}
	if (mincycles < DEF_info[0].TolerantTime + 2)
	{
		DoLog(LOG_GOALIE,"Safe Def");
		return GoalieRunto(defendpoint,0);
	}
	else
	{
		DoLog(LOG_GOALIE,"strive Def");
		if (fieldinfo.NumNoneGoalieTeammatesInCone(ballpoint,Defendline.GetPoint(Self.pos.dist(ballpoint)/3), Defendline.StrechAngle(fieldinfo.l_mygoalpost) *2) >= 1
			||Line(Defendline).dist(Self.pos) > SP_catch_area_w)
		{
			if (GoalieRunto(defendpoint,0))
			return true;
		}
		
		Vector prpoint = Defendline.GetClosestPoint(Self.pos);
		if (Defendline.DistanceFromOrigin(prpoint) < CP_safecatch_width)
			defendpoint = Defendline.GetPoint(CP_safecatch_width);
		else
			defendpoint = Defendline.GetPoint(Defendline.DistanceFromOrigin(prpoint)*2/3);
		if (!fieldinfo.my_penaltyarea.IsWithin(defendpoint))
				if	(!fieldinfo.my_penaltyarea.RayIntersection(Defendline,defendpoint))
					DoLog(LOG_BUG,"Why no intersection ?");
		DoLog(LOG_GOALIE,"OUT Dashing");
		return GoalieRunto(defendpoint,0);
		
	}
}

bool Goalie::DefendPass(){
	int i;
	Vector NewDefPos;
	
	float shoottimel = ballrollingtime(SP_ball_speed_max,ballpoint.dist(fieldinfo.l_mygoalpost));
	float shoottimer = ballrollingtime(SP_ball_speed_max,ballpoint.dist(fieldinfo.r_mygoalpost));
	float shoottime;
	if (BallInNarrowAngle())
	{
		shoottime = Min(shoottimel,shoottimer);
		float tarkang = (ballpoint - closerpost).Angle() + 90;
		if (fabs(tarkang) > 90) tarkang = NormalizeAngle(tarkang + 180);
		DEF_info[0].DefendBase = closerpost + Polar2Vector(Min(SP_goal_area_length+1.0f,SP_player_speed_max * (shoottime - 4)),tarkang);
		DEF_info[0].TolerantTime += shoottime/2;
		DoLog(LOG_GOALIE,"narrow ang new tol %.1f",DEF_info[0].TolerantTime);
	}
	else
	{
		shoottime =Min(shoottimel - Self.pos.dist(fieldinfo.l_mygoalpost) / SP_player_speed_max,
						shoottimer - Self.pos.dist(fieldinfo.r_mygoalpost) / SP_player_speed_max);
		DEF_info[0].TolerantTime += shoottime/2;
		if (DEF_info[0].TolerantTime == 0 )
		{
			DoLog(LOG_GOALIE,"No Time For pass def");
			return false;
		}
		if (!fieldinfo.my_goaldefendbox.RayIntersection(Defendline,DEF_info[0].DefendBase))
		DEF_info[0].DefendBase = Self.pos;
	}
	DoLog(LOG_GOALIE,"Basic Def at %.1f %.1f",DEF_info[0].DefendBase.x,DEF_info[0].DefendBase.y);
	Ray ConsiderLine;
	Ray ballray;
	Vector Ppos;
	Line pLine;
	for (i = 0; i < motion.Num_TheirVisiblePlayers();i ++)
	{
		if (num_opp_considered >= 5) break;
		if (motion.TheirPlayers_Danger[i] == motion.TheirPlayers_Interception[0]
			||situation.IsControllerOrFutureController(situation.TheirSide, motion.TheirPlayers_Danger[i])) continue;
		Ppos =TheirPlayer(motion.TheirPlayers_Danger[i]).pos;
		if (TheirPlayer(motion.TheirPlayers_Danger[i]).GetDefsensitivity() < 0.89f) break;
		//if i can intercept the ball, get ready to intercept;
		ballray.SetValues(ballpoint,(Ppos-ballpoint).Angle());
		Line pLine(Ray((Self.pos + Ppos)/2,(Self.pos - Ppos).Angle() + 90));
		if (pLine.HalfPlaneTest(Self.pos) == pLine.HalfPlaneTest(ballpoint) && GetGothroughSpeed(ballray,Self.pos,pLine) > SP_ball_speed_max
			&& fieldinfo.my_penaltyarea.IsWithin(Ppos))

		{
			DEF_info[num_opp_considered].OppPos = Ppos;
			NewDefPos = ballray.GetClosestPoint(Self.pos);
			DEF_info[num_opp_considered].DefendBase = DEF_info[0].DefendBase + 
				Polar2Vector(SP_player_speed_max * Min(DEF_info[0].TolerantTime/3,3.5f),(NewDefPos - Self.pos).Angle());
			DEF_info[num_opp_considered++].TolerantTime = Max(0.0f,ballrollingtime(SP_ball_speed_max,ballray.DistanceFromOrigin(NewDefPos))
				  - NewDefPos.dist(Self.pos) /SP_player_speed_max);
			DoLog(LOG_GOALIE,"Block # %u (%.1f,%.1f) at (%.1f,%.1f)",motion.TheirPlayers_Danger[i],Ppos.x,Ppos.y,DEF_info[num_opp_considered].DefendBase.x,DEF_info[num_opp_considered].DefendBase.y);
			continue;
		};
		//get to the balanced defend point
		ConsiderLine.SetValues(Ppos,MidAngle2((fieldinfo.l_mygoalpost - Ppos).Angle(),(fieldinfo.r_mygoalpost - Ppos).Angle())); 
			DEF_info[num_opp_considered].OppPos = Ppos;
			if (!fieldinfo.my_goaldefendbox.RayIntersection(ConsiderLine,NewDefPos))
			{
				DoLog(LOG_BUG,"why no intersection 2");
				continue;
			}
			CalcToleranceTime(DEF_info[num_opp_considered]);
			DEF_info[num_opp_considered ++].DefendBase = NewDefPos;
			DoLog(LOG_GOALIE,"Def # %u (%.1f,%.1f)",motion.TheirPlayers_Danger[i],Ppos.x,Ppos.y);
			
		

	}
	if (num_opp_considered ==1)
		return GoalieRunto(DEF_info[0].DefendBase,0);
	float weights[5];
	float sumweights = 0.0f;
	weights[0] = fieldinfo.GetDefensiveSensitivity(ballpoint);
	for (i=1;i<num_opp_considered;i++)
	{
		if (DEF_info[i].DefendBase.dist(DEF_info[0].DefendBase)
			< (DEF_info[i].TolerantTime + DEF_info[0].TolerantTime) * SP_player_speed_max)
		{
			weights[i] = fieldinfo.GetDefensiveSensitivity(DEF_info[i].OppPos);
			sumweights += weights[i];
			DEF_info[i].DefendBase = (DEF_info[0].DefendBase * weights[i] + DEF_info[i].DefendBase * weights[0])
				 /(weights[0] + weights[i]);
		}
		else
			weights[i] = 0.0f;
	}
	if (sumweights ==0.0f)
		return GoalieRunto(DEF_info[0].DefendBase,0);
	else
	{
		NewDefPos = Vector(0,0);
		for (i=1;i<num_opp_considered;i++)
			NewDefPos += DEF_info[i].DefendBase * (weights[i]/sumweights);
		return GoalieRunto(NewDefPos,0);
	}
}

void Goalie::AdjustFacings(){
	AngleDeg bodyfacing;
	Vector OppPos;
	if (fieldinfo.NumNoneGoalieTeammatesInCone(ballpoint,Defendline.GetPoint(3.0f),
		Defendline.StrechAngle(fieldinfo.l_mygoalpost) * 2) <= 0
		&&!(CheckShootSpeed(ballpoint + Polar2Vector(2.0f,(ballpoint - fieldinfo.l_mygoalpost).Angle()),Self.pos)
		  &&CheckShootSpeed(ballpoint,Self.pos) 
		  && CheckShootSpeed(ballpoint + Polar2Vector(2.0f,(ballpoint - fieldinfo.r_mygoalpost).Angle()),Self.pos)))
		//prepare to dash out 
	{
		DoLog(LOG_GOALIE,"prep dash");
		Vector prpoint = Defendline.GetClosestPoint(Self.pos);
		if (Defendline.DistanceFromOrigin(prpoint) < CP_safecatch_width)
			prpoint = Defendline.GetPoint(CP_safecatch_width);
		else
			prpoint = Defendline.GetPoint(Defendline.DistanceFromOrigin(prpoint)*2/3);
	
		bodyfacing = motion.turn_to(prpoint);
		nextfacing = Self.bodyfacing + bodyfacing;
	}
	else
	{
		if (num_opp_considered <=1 )
			OppPos = ballpoint;
		else
			OppPos = fieldinfo.GetDefensiveSensitivity(ballpoint) > fieldinfo.GetDefensiveSensitivity(DEF_info[1].OppPos) ?
					ballpoint : DEF_info[1].OppPos;
		DoLog(LOG_GOALIE,"advise %.1f,%.1f",OppPos.x,OppPos.y);
		Line DOrgLine;
		Command command;
		DOrgLine.LineFromTwoPoints(OppPos,fieldinfo.mygoal);
		if(BallInNarrowAngle())
		{
			if (ball.pos.y > 0) bodyfacing = 90;
			else
				bodyfacing = - 90;
		}
		else
		{
			Vector defpoint = DOrgLine.ProjectPoint(Self.pos);
			if (!fieldinfo.WithInField(defpoint))
				defpoint = DOrgLine.intersection(bottomline);
			bodyfacing = (defpoint - Self.pos).Angle();
			if (fabs(NormalizeAngle(bodyfacing - (defpoint - fieldinfo.mygoal).Angle())) > 90)
			{
				bodyfacing = NormalizeAngle(bodyfacing + 180);
			}
		}
		bodyfacing = Self.bodyfacing + motion.turn_to(bodyfacing,command);
		// keep ball in focus
		if (fabs(NormalizeAngle(bodyfacing - ball.global_angle)) > SP_max_neck_angle)
			bodyfacing = fabs(NormalizeAngle(bodyfacing - ball.global_angle + SP_max_neck_angle))
						 > fabs(NormalizeAngle(bodyfacing - ball.global_angle - SP_max_neck_angle))
						 ? ball.global_angle + SP_max_neck_angle : ball.global_angle - SP_max_neck_angle;
			motion.turn_to(bodyfacing);
		nextfacing = bodyfacing;
	}
}

bool Goalie::GoalieRunto(Vector point ,int deviationdegree,float speed){
	Command command;
	DoLog(LOG_GOALIE,"target %.1f,%.1f",point.x,point.y);
	if (point.x < CP_safecatch_width - SP_semi_pitch_length)
	{
		point.x = CP_safecatch_width - SP_semi_pitch_length;
		DoLog(LOG_GOALIE,"Adjust posx %.1f",point.x);
	}	
	Vector gap = point - Self.PredictPos();
	float angle;
	
	if (deviationdegree == 0)
	{
		if (gap.mod() < ball.distance / 20)
		{
			DoLog(LOG_GOALIE,"Already reach");
			return false;
		}
			
		//perform a smooth motion
		if (speed > 0)
			speed = Min(speed,(gap + Self.global_vel).mod() * (1- SP_player_decay));
		else
			speed = Max(speed,-(gap + Self.global_vel).mod() * (1- SP_player_decay));
		
		if (speed < 0)
			angle = NormalizeAngle(gap.Angle() + 180);
		else
			angle = gap.Angle();

		//should dash backward to keep correct facing;	
		if (ball.pos_valid() &&(speed >0 && fabs(NormalizeAngle(gap.Angle() - ball.global_angle)) >90
			|| speed < 0 && fabs(NormalizeAngle(gap.Angle() - ball.global_angle)) < 90))
		{
			speed = - speed;
			angle = NormalizeAngle(angle + 180);
		}	 
	if (fabs(Self.pos.dist(point) * Sin(angle - Self.bodyfacing)) > SP_kickable_margin && 
		fabs(NormalizeAngle(angle -Self.bodyfacing)) > 15
		|| fabs(NormalizeAngle(angle - Self.bodyfacing)) > 90)
	{
			angle = motion.turn_to(angle, command);
			nextfacing = Self.bodyfacing + angle;
		}
		else{
			motion.run(speed, command);
			nextfacing = Self.bodyfacing;
		}
		mediator.enroll(command,Action_positioning_defense,PriorityA * 5.0f); 
		return true;
	}
	
	float controlmargin;
	if (deviationdegree ==2)
		controlmargin = CP_reliable_kickarea * 0.95f;
	else
		controlmargin = SP_catch_area_l;
	if (speed < 0)
			angle = NormalizeAngle(gap.Angle() + 180);
		else
			angle = NormalizeAngle(gap.Angle());

	if (fabs(Self.pos.dist(point) * Sin(angle - Self.bodyfacing)) > controlmargin
		|| fabs(NormalizeAngle(angle - Self.bodyfacing)) > 90)
	{
		angle = motion.turn_to(angle,command);
		nextfacing = Self.bodyfacing + angle;
	}
	else
	{
		motion.run(speed,command);
		nextfacing = Self.bodyfacing;
	}
		mediator.enroll(command,Action_interception,PriorityA);
	return true;
	
} 

void Goalie::Catchball(){
	action.seize(NormalizeAngle(ball.global_angle - Self.bodyfacing));
}

void Goalie::Kickaway(){
	kick.smartkick(SP_ball_speed_max, motion.GetClearAngle(),KK_fast); 
}

bool Goalie::IsShoot(){
	Ray ballray;
	if(ball.global_vel.x>=0) return false; 
	if (!situation.BallFree) 
	{
		DoLog(LOG_GOALIE,"Not ball free");
		return false;	
	}
	ballray.SetValues(ball.pos,ball.global_vel);
	Vector p;
	if(!fieldinfo.SideLines[SL_Left].RayIntersection(ballray,p))
	{
		DoLog(LOG_GOALIE,"No inters");
		return false;
	}
	if ( p == Vector(0,0) )
	{
		DoLog(LOG_GOALIE,"invalid inters");
		return false;
	}
	if ( (p.y > SP_goal_area_width /2) || (p.y < -SP_goal_area_width / 2) ) 
	{
		DoLog(LOG_GOALIE,"OUT  %.1f %.1f",p.x,p.y);
		return false;
	}
	bottomline.RayIntersection(ballray,p);
	float maxballdist=ball.speed/(1-SP_ball_decay);
	return bool(p.dist(ball.pos)<maxballdist);
}

void Goalie::AddPITinfo(PassiveITInfo& PassiveIT_info){
	PIT_info[num_PIT_infos]=PassiveIT_info;
	num_PIT_infos++;
	DoLog(LOG_GOALIE,"PIT cy %.1f, sp %.1f,wt %.1f,at %.1f %.1f",PassiveIT_info.IT_cycles,
		PassiveIT_info.IT_rapidness,PassiveIT_info.IT_waitcycles,PassiveIT_info.IT_point.x,PassiveIT_info.IT_point.y);
} 

bool Goalie::GetInterceptionInfo(Vector ballpos,Vector ballvel,Vector Selfpos,Vector Selfvel,AngleDeg facing){
	mostpromisingIT.turnangle = InvalidAngle;
	float mingap = SP_pitch_diameter;
	PassiveITInfo IT_info;
	int Turncycle;
	float angdif;
	float controlmargin;
	if (!fieldinfo.WithInField(ballpos)) return false;
	Vector orgballpos = ballpos;
	
	const int CP_stational_cycles = 6;
	Vector playerpos[CP_stational_cycles + 1];
	Vector playervel[CP_stational_cycles + 1];
	playerpos[0] = Selfpos;
	playervel[0] = Selfvel;
	int i;
	int reachballcycle;
	for (i=1;i <= CP_stational_cycles;i++)
	{
		playerpos[i] = playerpos[i-1] + playervel[i-1];
		playervel[i] = playervel[i] * SP_player_decay;
	}
	
	if (ballvel.mod() < 0.10f)
	{
		if (fieldinfo.my_penaltyarea.IsWithin(ballpos))
			controlmargin = SP_catch_area_l;
		else
			controlmargin = SP_kickable_area;
		controlmargin *= 0.95f;
		IT_info.IT_point= ballpos;
		IT_info.IT_waitcycles = 0;
		IT_info.turnangle = (ballpos - Selfpos).Angle();
		angdif = float(fabs(NormalizeAngle(IT_info.turnangle - facing)));
		Turncycle = 0;
		while (ballpos.dist(playerpos[Turncycle]) * Sin(angdif) > controlmargin){
				angdif -=action.Turnrate(playervel[Turncycle].mod()) * SP_max_moment;
				if (angdif <0) angdif =0;
				Turncycle ++;
			}
		IT_info.IT_rapidness = Min(SP_player_speed_max,(ballpos - Selfpos).mod());
		IT_info.IT_cycles =(ballpos - Selfpos).mod()  / SP_player_speed_max + Turncycle;
		AddPITinfo(IT_info);
		if (angdif >90 && ball.distance < 5 * SP_player_speed_max)
		{
			angdif = 180 - angdif;
			Turncycle = 0;
			while (ballpos.dist(playerpos[Turncycle]) * Sin(angdif) > controlmargin){
					angdif -=action.Turnrate(playervel[Turncycle].mod()) * SP_max_moment;
					if (angdif <0) angdif =0;
					Turncycle ++;
				}
			DoLog(LOG_GOALIE,"Turncycle %d",Turncycle);
			IT_info.IT_rapidness = Max(- SP_player_speed_max,- ballpos.dist(Selfpos));
			IT_info.IT_cycles = ballpos.dist(Selfpos) / SP_player_speed_max + Turncycle;		
			AddPITinfo(IT_info);
		}
		return true;
	}

	Ray ballcourse;
	ballcourse.SetValues(ballpos,ballvel);
	Vector closestpoint = ballcourse.GetClosestPoint(Selfpos);
	if (!fieldinfo.WithInField(closestpoint))
	ballcourse.intersection(fieldinfo.SideLines[SL_Left],closestpoint);
	float rundist = Selfpos.dist(closestpoint);
	float mincycle = (rundist - SP_catch_area_l)/ SP_player_speed_max;
	if (mincycle <0) mincycle = 0;
	Vector p1;
	while (ballrollingtime(ballvel.mod(),ballcourse.DistanceFromOrigin(closestpoint)) < mincycle)
	{
		rundist += SP_player_speed_max;
		mincycle +=1;
		if (ballcourse.CircleIntersect(rundist,Selfpos,p1,closestpoint) == 0)
			return false;
		if (!fieldinfo.WithInField(closestpoint))
			return false;

	}
	

	int ITCycle = int(mincycle);
	int FullDashCycle;
	bool canbackdash;
	int CP_back_dash_max_cycles = 15;
	Vector shadowpoint,dashtopoint,dashvel;
	DoLog(LOG_GOALIE,"Min by projectpoint %d",ITCycle);
	proceedball(ballpos,ballvel,ITCycle);
	do{
		FullDashCycle = Min(CP_stational_cycles,ITCycle);
		if (fieldinfo.my_penaltyarea.IsWithin(ballpos))
			controlmargin = SP_catch_area_l;
		else
			controlmargin = SP_kickable_area;
		IT_info.turnangle = (ballpos - playerpos[FullDashCycle]).Angle();
		
		if ((ballpos - playerpos[FullDashCycle]).mod() < controlmargin)
		{
			IT_info.IT_cycles = (float)ITCycle;
			IT_info.IT_point= ballpos;
			IT_info.IT_rapidness = Min((ballpos - playerpos[FullDashCycle]).mod(),SP_player_speed_max);
			IT_info.IT_waitcycles =float(ITCycle);
			AddPITinfo(IT_info);			
		}
		
		if (ITCycle > CP_back_dash_max_cycles)
			canbackdash = false;
		else canbackdash = true;
		// check if back dash will lost sight of ball
		if (canbackdash && ITCycle > 5)
		{
			shadowpoint = ballpos - ballvel * 6.0f;	    
			if ((shadowpoint - Selfpos).mod() > SP_player_speed_max * (ITCycle - 5) + 3.0f
				&&ballcourse.DistanceFromOrigin(ballcourse.GetClosestPoint(Selfpos)) < 5 * ballvel.mod())
			{
				
				dashtopoint = ballpos - Polar2Vector(5 * SP_player_speed_max + controlmargin,IT_info.turnangle);
				if (fabs(NormalizeAngle(IT_info.turnangle - (shadowpoint - dashtopoint).Angle())) > 45)
				canbackdash = false; 
			}
		}
		//check if turn and dashing forward can get the ball
		angdif = (float)fabs(NormalizeAngle(IT_info.turnangle - facing));
		if (angdif < 90) canbackdash = false;
		Turncycle = 0;
		while (ballpos.dist(playerpos[Turncycle]) * Sin(angdif) > controlmargin || (angdif > 90)){
			angdif -=action.Turnrate(playervel[Turncycle].mod()) * SP_max_moment;
			if (angdif <0) angdif =0;
			Turncycle ++;
		}
		reachballcycle = -1;
		if (Turncycle <= ITCycle)
		{
			dashtopoint = playerpos[Turncycle];
			dashvel = playervel[Turncycle];
			for (i = Turncycle; i < FullDashCycle;i++)
			{
				dashvel +=Polar2Vector(SP_max_power * action.Dashrate(),IT_info.turnangle);
				if (dashvel.mod() > SP_player_speed_max)
					dashvel =dashvel/dashvel.mod() * SP_player_speed_max;
				dashtopoint += dashvel;
				if (dashtopoint.dist(ballpos) < controlmargin)
					reachballcycle = i+1;
			}
			if ((dashtopoint.dist(ballpos) - controlmargin) /SP_player_speed_max 
				<= ITCycle - FullDashCycle)
			{
				IT_info.IT_cycles = (float)ITCycle;
				IT_info.IT_point= ballpos;
				if (angdif < 5 && ballvel.mod() > 1.4f)
					IT_info.IT_point+= Polar2Vector(SP_kickable_area /2,ballvel.Angle());
				IT_info.IT_rapidness = Min(SP_player_speed_max,(float)fabs(orgballpos.dist(Self.PredictPos(Turncycle)) * Cos(angdif)));
				if (reachballcycle != -1)
					IT_info.IT_waitcycles =float(ITCycle - reachballcycle);
				else
					IT_info.IT_waitcycles =ITCycle - FullDashCycle - Max((dashtopoint.dist(ballpos) - controlmargin) /SP_player_speed_max , 0.0f);
				AddPITinfo(IT_info);
			}
			else
			{
				if (ITCycle >0)
				if (dashtopoint.dist(ballpos)/ITCycle < mingap)
				{
					mingap = dashtopoint.dist(ballpos)/ITCycle;
					mostpromisingIT.turnangle = IT_info.turnangle;
					mostpromisingIT.IT_rapidness = SP_player_speed_max;
					mostpromisingIT.IT_point= ballpos;
				}
			}

		}
		//check if turn and dash backward can get the ball
		if (canbackdash)
		{
			angdif = float(180 - fabs(NormalizeAngle(IT_info.turnangle - facing)));
			Turncycle = 0;
			while (ballpos.dist(playerpos[Turncycle]) * Sin(angdif) > controlmargin || angdif > 90 ){
				angdif -=action.Turnrate(playervel[Turncycle].mod()) * SP_max_moment;
				if (angdif <0) angdif =0;
				Turncycle ++;
			}
			if (Turncycle <= ITCycle)
			{
				dashtopoint = playerpos[Turncycle];
				dashvel = playervel[Turncycle];
				reachballcycle = -1;
				for (i = Turncycle; i < FullDashCycle;i++)
				{
					dashvel +=Polar2Vector(SP_max_power * action.Dashrate(),IT_info.turnangle);
					if (dashvel.mod() > SP_player_speed_max)
						dashvel =dashvel/dashvel.mod() * SP_player_speed_max;
					dashtopoint += dashvel;
					if (dashtopoint.dist(ballpos)<controlmargin)
						reachballcycle = i+1;
				}
				if ((dashtopoint.dist(ballpos) - controlmargin) /SP_player_speed_max 
					<= ITCycle - FullDashCycle)
				{
					IT_info.IT_cycles = (float)ITCycle;
					IT_info.IT_point= ballpos;
					if (angdif < 5 && ballvel.mod() > 1.4f)
						IT_info.IT_point+= Polar2Vector(SP_kickable_area /2,ballvel.Angle());
					IT_info.IT_rapidness = Max( -SP_player_speed_max,- (float)fabs(orgballpos.dist(Self.PredictPos(Turncycle)) * Cos (angdif)));
					if (reachballcycle != -1)
						IT_info.IT_waitcycles =float(ITCycle - reachballcycle);
					else
						IT_info.IT_waitcycles =ITCycle - FullDashCycle - Max((dashtopoint.dist(ballpos) - controlmargin) /SP_player_speed_max,0.0f);
					AddPITinfo(IT_info);
				}
				else
				{
					if (ITCycle > 0)
					if (dashtopoint.dist(ballpos)/ITCycle < mingap)
					{
						mingap = dashtopoint.dist(ballpos)/ITCycle;
						mostpromisingIT.turnangle = IT_info.turnangle;
						mostpromisingIT.IT_rapidness = SP_player_speed_max;
						mostpromisingIT.IT_point= ballpos;
					}
				}
			}
		}
		ballpos +=ballvel;
		ballvel *=SP_ball_decay;
		ITCycle ++;
	}while (fieldinfo.WithInField(ballpos) && (ballvel.mod() > 0.05f * ITCycle));
	if (num_PIT_infos ==0 && Self.IT_inf.IT_cycles >ITCycle && Self.IT_inf.IT_cycles < InfCycles
		&& fieldinfo.WithInField(Self.IT_inf.IT_point))
	{
		IT_info.IT_cycles = Self.IT_inf.IT_cycles;
		IT_info.IT_rapidness = Self.IT_inf.IT_rapidness;
		IT_info.turnangle = Self.IT_inf.IT_angle;
		IT_info.IT_point= Self.IT_inf.IT_point;
		IT_info.IT_waitcycles = -1;//by net
		AddPITinfo(IT_info);

	}
	return (num_PIT_infos > 0);
}


bool Goalie::OneCycleIT(){ 
	if (!fieldinfo.my_penaltyarea.IsWithin(ball.PredictPos(1)) && motion.TheirPlayer_Close2Ball(0).balldist > 5.0f)
		return false;
	Vector ballrelpos = ball.PredictPos(1) - Self.PredictPos() ;
	float dashpow = 2 * SP_max_power;
	ballrelpos = ballrelpos.Rotate( - Self.bodyfacing);
	if (fabs(ballrelpos.y) > SP_catch_area_l) return false;
	Vector ballvel = ball.global_vel;
	Vector next_ball_pos;
	float maxdasheffect = action.Max_DashEffect();
	if (maxdasheffect > fabs(ballrelpos.x)){
		if (fabs(ballrelpos.y) >= CP_desired_kickarea){
			dashpow = float(fabs(ballrelpos.x) / action.Dashrate());
			DoLog(LOG_GOALIE, " One Cyc IT1");
		}
		else{
		//There are two solutions, choose the one that is more robust
			float solution[2], dist[2];
			int final_solution = 0;
			solution[0] = float(fabs(ballrelpos.x) - sqrt(Sqr(CP_desired_kickarea) - Sqr(ballrelpos.y)));
			solution[1] = 2 * (float)fabs(ballrelpos.x) - solution[0];
			pose_limitation(solution[0], -action.Max_DashEffect(), action.Max_DashEffect());
			solution[1] = Min(solution[1], action.Max_DashEffect());
			dist[0] = float(sqrt(Sqr(solution[0] - fabs(ballrelpos.x)) + Sqr(ballrelpos.y)));
			dist[1] = float(sqrt(Sqr(solution[1] - fabs(ballrelpos.x)) + Sqr(ballrelpos.y)));
			if (dist[0] < CP_collision_threshold || dist[1] < CP_collision_threshold){
				//if one solution leads to a collision, choose the bigger one
				if (dist[0] > dist[1])
					final_solution = 0;
				else
					final_solution = 1;
			}
			else{
			//try to intercept with front side
				ballvel.Rotate(-Self.bodyfacing);
				next_ball_pos = ballvel + ballrelpos;
				if (Vector(solution[0], 0).dist2(next_ball_pos) > Vector(solution[1], 0).dist2(next_ball_pos))
					final_solution = 1;
				else
					final_solution =0;
			}
			
			dashpow = solution[final_solution] / action.Dashrate();
			DoLog(LOG_GOALIE, "One cyc IT2");
		}
	}
	else{
		float min_dist = (float)sqrt(Sqr(fabs(ballrelpos.x) - maxdasheffect) + Sqr(ballrelpos.y));
		float GoalieIT_area = fieldinfo.my_penaltyarea.IsWithin(Self.PredictPos(SP_max_power)) ? SP_catch_area_l : SP_kickable_area;

		if (min_dist < GoalieIT_area){
			if (min_dist >= CP_desired_kickarea){
				dashpow = SP_max_power;
			}
			else{
				dashpow = Max(float(fabs(ballrelpos.x) - sqrt(Sqr(CP_desired_kickarea) - Sqr(ballrelpos.y))) / action.Dashrate(), -SP_max_power);
			}
			DoLog(LOG_GOALIE,"One cyc IT 3");
		}
	}

	if (dashpow == 2 * SP_max_power) return false;
	if (ballrelpos.x < 0) dashpow *= -1.0f;
	Command command;
	command.dash(dashpow);
	mediator.enroll(command,Action_interception,2 * PriorityA);
	return true;
	
}


bool Goalie::GoalieIntercept(){
	int i;
	bool haveintercept = false;
	if (OneCycleIT()) 
	{
		DoLog(LOG_GOALIE,"One Cycle and get ball");
		haveintercept = true;
	}
	if (!ball.pos_valid()) return  false;
	DoLog(LOG_GOALIE,"Try intercept");
	int bestPIT = -1;
	if (num_PIT_infos == 0)
	{
		if (!IsShoot()|| IfOppkickmore()) 
		{
			DoLog(LOG_GOALIE,"Not final shooting");
			return false;
		}
		DoLog(LOG_GOALIE,"IS Shoot so Try");
		minITcycles = 0;
		if (mostpromisingIT.turnangle != InvalidAngle)
		{
			GoalieRunto(mostpromisingIT.IT_point,1);
			haveintercept = true;
		}
		else
		{
			Vector crosspoint;
			Ray ballray;
			ballray.SetValues(ball.pos,ball.global_vel);
			if	(!bottomline.RayIntersection(ballray,crosspoint))
			{
				haveintercept = GoalieRunto(ball.PredictPos(1),1);
				return haveintercept;
			}
			GoalieRunto(crosspoint,1);
			haveintercept = true;
		}		
	}
	else
	{
		float oppmincycles;
		if (motion.Num_TheirVisiblePlayers() == 0)
			oppmincycles = InfCycles;
		else
		oppmincycles= (situation.BallFree || !IsTheirPlayer(situation.ballcontroller)) ? TheirPlayer(motion.TheirPlayers_Interception[0]).IT_inf.IT_cycles : 3.0f;
		DoLog(LOG_GOALIE,"Their cycles %.1f",oppmincycles);
		int dev_degree;
		while (num_PIT_infos > 0 && PIT_info[num_PIT_infos -1].IT_cycles > 2 && (PIT_info[num_PIT_infos -1].IT_cycles > oppmincycles
			||PIT_info[num_PIT_infos -1].IT_cycles > oppmincycles * 0.9f && !fieldinfo.my_penaltyarea.IsWithin(PIT_info[num_PIT_infos -1].IT_point)))
			num_PIT_infos --;

		float mycycles;
		if (!situation.BallFree)
		{
			if (IsMyPlayer(situation.ballcontroller))
				mycycles = 0.0f;
			else
				mycycles = 3.0f;
		}
		else
		if (motion.MyPlayers_Interception[0] ==FP_goalie_number)
		{
			if (motion.Num_MyVisiblePlayers() <1)
				mycycles = InfCycles;
			else
				mycycles = motion.MyPlayer_CtrlOfBall(1).IT_inf.IT_cycles;
		}
		else
			mycycles = motion.MyPlayer_CtrlOfBall(0).IT_inf.IT_cycles;
		
		while (num_PIT_infos > 0 &&  (PIT_info[num_PIT_infos -1].IT_cycles > mycycles + 2.0f
			||PIT_info[num_PIT_infos -1].IT_cycles > mycycles  * 0.9f && !fieldinfo.my_penaltyarea.IsWithin(PIT_info[num_PIT_infos -1].IT_point)))
			num_PIT_infos --;
		if (motion.calculatecontrol(mycycles,oppmincycles) > 0.6f) 
		{
			while (num_PIT_infos >0 &&!fieldinfo.my_penaltyarea.IsWithin(PIT_info[num_PIT_infos -1].IT_point))
				num_PIT_infos --;
		}
		
		if (num_PIT_infos == 0)
			return false;
		
		if (PIT_info[0].IT_cycles > oppmincycles -2)
		{
			if (fieldinfo.my_penaltyarea.IsWithin(PIT_info[0].IT_point))
				dev_degree = 1;
			else
				dev_degree = 2;
			bestPIT = 0;
			GoalieRunto(PIT_info[0].IT_point,dev_degree,PIT_info[0].IT_rapidness);
			haveintercept = true;
			DoLog(LOG_GOALIE,"My cycle %.1f",PIT_info[0].IT_cycles);
		}
		else
		{
			while (num_PIT_infos > 0 && PIT_info[num_PIT_infos -1].IT_cycles > oppmincycles -2)
				num_PIT_infos --;
			int  first_considered_PIT = 0;
			if (fieldinfo.my_penaltyarea.IsWithin(PIT_info[num_PIT_infos -1].IT_point))
				while (!fieldinfo.my_penaltyarea.IsWithin(PIT_info[first_considered_PIT].IT_point) &&first_considered_PIT < num_PIT_infos -1)
						first_considered_PIT ++;
			bestPIT = first_considered_PIT;
			for (i = first_considered_PIT ;i <num_PIT_infos; i++)
				if (PIT_info[i].IT_waitcycles >=1.0f && PIT_info[i].IT_rapidness >0 &&
					PIT_info[i].IT_point.x > -SP_semi_pitch_length - CP_safecatch_width)
				{
					bestPIT =i;
					break;
				}
			if (bestPIT == first_considered_PIT)
			for (i =first_considered_PIT ;i <num_PIT_infos; i++)
				if (PIT_info[i].IT_waitcycles >=1.0f &&
					PIT_info[i].IT_point.x > -SP_semi_pitch_length - CP_safecatch_width)
				{
					bestPIT =i;
					break;
				}
			if (fieldinfo.my_penaltyarea.IsWithin(PIT_info[bestPIT].IT_point))
				dev_degree = 1;
			else
				dev_degree = 2;
			if (!GoalieRunto(PIT_info[bestPIT].IT_point,dev_degree,PIT_info[bestPIT].IT_rapidness))
				motion.turn_to(ball.PredictPos(1));
			haveintercept = true;
			DoLog(LOG_GOALIE,"best %d My cycle  %.1f",bestPIT,PIT_info[bestPIT].IT_cycles);
		}
	}
	VIEWWIDTH myviewwidth;
	float ballfreecycles = Min(minITcycles,TheirPlayer(motion.TheirPlayers_Interception[0]).IT_inf.IT_cycles);
	
	if (ballfreecycles < 4)
		myviewwidth = VW_Narrow;
	else if (ballfreecycles < 10 )
		myviewwidth = VW_Normal;
	else
		myviewwidth = VW_Wide;
	if (bestPIT != -1)
	{
		if (PIT_info[bestPIT].IT_waitcycles < 5 && myviewwidth == VW_Wide)
			myviewwidth = VW_Normal;
	}
	
	return haveintercept;
}

Vector Goalie::GetGuardPoint(){

	return Vector(10.0f - SP_semi_pitch_length ,0);
} 

float Goalie::ballrollingtime(float orgspeed,float dist){
	if (orgspeed < 0.15f) return InfCycles;
	if (dist >orgspeed / (1-SP_ball_decay)) return InfCycles;
	return float (log(1.0f - (dist/orgspeed)*(1.0f - SP_ball_decay))/log(SP_ball_decay));
}

float Goalie::ballspeedrequired(float dist,int time){
	return float(dist * (1 - SP_ball_decay) /(1 - Exp(SP_ball_decay,(float)time)));
}

void Goalie::CalcToleranceTime(DefendInfo &definf){
	float cycles = 0;
	if (situation.BallFree) 
		cycles += situation.min_IT_cycles;
	definf.TolerantTime = cycles;
	if (ballpoint.dist(definf.OppPos) > CP_different_distance)
	definf.TolerantTime += ballrollingtime(SP_ball_speed_max + SP_player_speed_max,ballpoint.dist(definf.OppPos));
}

float Goalie::dist_to_goal(Vector pos){
	if (fabs(pos.y) <= SP_goal_width /2)
		return (pos.x + SP_semi_pitch_length);
	if (pos.y > 0) return pos.dist(fieldinfo.r_mygoalpost);
	else return pos.dist(fieldinfo.l_mygoalpost);
}

void Goalie::proceedball(Vector &ballpos,Vector &ballvel,int cycles){
	for (int i=0;i < cycles;i++)
	{
		ballpos += ballvel;
		ballvel *=SP_ball_decay;
	}
}

bool Goalie::BallInNarrowAngle(){ 
	if (fabs(ballpoint.y) < SP_goal_width) return false;
	Vector farpost = ballpoint.y >0 ? fieldinfo.l_mygoalpost :fieldinfo.r_mygoalpost;	
	return fabs(SP_goal_width * Cos((ballpoint - farpost).Angle())) <= 2.5 * SP_catch_area_l;
}

bool Goalie::CatchWait(){
	
	if (!situation.BallFree) return false;
	if (ball.distance < SP_catch_area_l * 0.85f) return false;
	if (!fieldinfo.WithInField(ball.PredictPos(1)) ||ball.PredictPos(1).x < CP_safecatch_width - SP_semi_pitch_length)
		return false;
	Vector Mydashend1 = Self.PredictPos(SP_max_power);
	Vector Mydashend2 = Self.PredictPos( -SP_min_power);
	Line Mydashline;
	Mydashline.LineFromTwoPoints(Mydashend1,Mydashend2);
	Vector pointpr = Mydashline.ProjectPoint(ball.PredictPos(1));
	float mindis;
	if (Mydashline.InBetween(pointpr,Mydashend1,Mydashend2))
		mindis = pointpr.dist(ball.PredictPos(1));
	else
		mindis = Min(ball.PredictPos(1).dist(Mydashend1),ball.PredictPos(2).dist(Mydashend2));
	return mindis < ball.distance * 0.9;
}